# getbinpkg.py -- Portage binary-package helper functions
# Copyright 2003-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.output import colorize
from portage.cache.mappings import slot_dict_class
from portage.localization import _
import portage
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.package.ebuild.fetch import _hide_url_passwd
from _emerge.Package import _all_metadata_keys

import pickle
import sys
import socket
import time
import tempfile
import base64
import warnings

from html.parser import HTMLParser as html_parser_HTMLParser
from urllib.parse import unquote as urllib_parse_unquote

_all_errors = [NotImplementedError, ValueError, socket.error]

try:
    import ftplib
except ImportError as e:
    sys.stderr.write(colorize("BAD", "!!! CANNOT IMPORT FTPLIB: ") + str(e) + "\n")
else:
    _all_errors.extend(ftplib.all_errors)

try:
    from http.client import HTTPConnection as http_client_HTTPConnection
    from http.client import BadStatusLine as http_client_BadStatusLine
    from http.client import ResponseNotReady as http_client_ResponseNotReady
    from http.client import error as http_client_error
except ImportError as e:
    sys.stderr.write(colorize("BAD", "!!! CANNOT IMPORT HTTP.CLIENT: ") + str(e) + "\n")
else:
    _all_errors.append(http_client_error)

_all_errors = tuple(_all_errors)


def make_metadata_dict(data):
    warnings.warn(
        "portage.getbinpkg.make_metadata_dict() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    myid, _myglob = data
    metadata = (
        (
            k_bytes,
            _unicode_decode(
                k_bytes, encoding=_encodings["repo.content"], errors="replace"
            ),
        )
        for k_bytes in portage.xpak.getindex_mem(myid)
    )
    mydict = {
        k: _unicode_decode(
            portage.xpak.getitem(data, k_bytes),
            encoding=_encodings["repo.content"],
            errors="replace",
        )
        for k_bytes, k in metadata
        if k in _all_metadata_keys or k == "CATEGORY"
    }

    return mydict


class ParseLinks(html_parser_HTMLParser):
    """Parser class that overrides HTMLParser to grab all anchors from an html
    page and provide suffix and prefix limitors"""

    def __init__(self):
        warnings.warn(
            "portage.getbinpkg.ParseLinks is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )

        self.PL_anchors = []
        html_parser_HTMLParser.__init__(self)

    def get_anchors(self):
        return self.PL_anchors

    def get_anchors_by_prefix(self, prefix):
        newlist = [
            x for x in self.PL_anchors if x.startswith(prefix) and x not in newlist
        ]
        return newlist

    def get_anchors_by_suffix(self, suffix):
        newlist = [
            x for x in self.PL_anchors if x.endswith(suffix) and x not in newlist
        ]
        return newlist

    def handle_endtag(self, tag):
        pass

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            myarchors = (
                urllib_parse_unquote(x[1])
                for x in attrs
                if x[0] == "href" and x[1] not in self.PL_anchors
            )
            self.PL_anchors.extend(myarchors)


def create_conn(baseurl, conn=None):
    """Takes a protocol://site:port/address url, and an
    optional connection. If connection is already active, it is passed on.
    baseurl is reduced to address and is returned in tuple (conn,address)"""

    warnings.warn(
        "portage.getbinpkg.create_conn() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    parts = baseurl.split("://", 1)
    if len(parts) != 2:
        raise ValueError(
            _("Provided URI does not " "contain protocol identifier. '%s'") % baseurl
        )
    protocol, url = parts
    del parts

    url_split = url.split("/", 1)
    host = url_split[0]
    address = f"/{url_split[1]}"
    del url, url_split

    userpass_host = host.split("@", 1)
    if len(userpass_host) == 1:
        host = userpass_host[0]
        userpass = ["anonymous"]
    else:
        host = userpass_host[1]
        userpass = userpass_host[0].split(":")
    del userpass_host

    userpass_size = len(userpass)
    if userpass_size > 2:
        raise ValueError(_("Unable to interpret username/password provided."))

    username = userpass[0]
    if userpass_size == 2:
        password = userpass[1]
    elif userpass_size == 1:
        password = None
    del userpass

    http_headers = {}
    http_params = {}
    if username and password:
        try:
            encodebytes = base64.encodebytes
        except AttributeError:
            # Python 2
            encodebytes = base64.encodestring
        unicode_bytes = encodebytes(_unicode_encode(f"{username}:{password}")).replace(
            b"\012", b""
        )
        http_headers = {b"Authorization": f"Basic {unicode_bytes}"}

    if not conn:
        if protocol == "https":
            # Use local import since https typically isn't needed, and
            # this way we can usually avoid triggering the global scope
            # http.client ImportError handler (like during stage1 -> stage2
            # builds where USE=ssl is disabled for python).
            try:
                from http.client import HTTPSConnection as http_client_HTTPSConnection
            except ImportError:
                raise NotImplementedError(
                    _("python must have ssl enabled for https support")
                )
            conn = http_client_HTTPSConnection(host)
        elif protocol == "http":
            conn = http_client_HTTPConnection(host)
        elif protocol == "ftp":
            passive = 1
            if host[-1] == "*":
                passive = 0
                host = host[:-1]
            conn = ftplib.FTP(host)
            if password:
                conn.login(username, password)
            else:
                sys.stderr.write(
                    colorize(
                        "WARN",
                        _(f" * No password provided for username '{username}'\n\n"),
                    )
                )
                conn.login(username)
            conn.set_pasv(passive)
            conn.set_debuglevel(0)
        elif protocol == "sftp":
            try:
                import paramiko
            except ImportError:
                raise NotImplementedError(
                    _("paramiko must be installed for sftp support")
                )
            t = paramiko.Transport(host)
            t.connect(username=username, password=password)
            conn = paramiko.SFTPClient.from_transport(t)
        else:
            raise NotImplementedError(_("%s is not a supported protocol.") % protocol)

    return (conn, protocol, address, http_params, http_headers)


def make_ftp_request(conn, address, rest=None, dest=None):
    """Uses the |conn| object to request the data
    from address and issuing a rest if it is passed."""

    warnings.warn(
        "portage.getbinpkg.make_ftp_request() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    try:
        if dest:
            fstart_pos = dest.tell()

        conn.voidcmd("TYPE I")
        fsize = conn.size(address)

        retr_address = f"RETR {address}"
        if rest and rest < 0:
            rest = fsize + int(rest)
            mysocket = conn.transfercmd(retr_address, rest)
        else:
            mysocket = conn.transfercmd(retr_address)

        mydata = ""
        while 1:
            somedata = mysocket.recv(8192)
            if somedata:
                if dest:
                    dest.write(somedata)
                else:
                    mydata = f"{mydata}{somedata}"
            else:
                break

        data_size = len(mydata)
        if dest:
            data_size = fstart_pos - dest.tell()

        mysocket.close()
        conn.voidresp()
        conn.voidcmd("TYPE A")

        return mydata, (fsize != data_size), ""

    except ValueError as e:
        return None, int(str(e)[:4]), str(e)


def make_http_request(conn, address, _params={}, headers={}, dest=None):
    """Uses the |conn| object to request
    the data from address, performing Location forwarding and using the
    optional params and headers."""

    warnings.warn(
        "portage.getbinpkg.make_http_request() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    rc = 0
    response = None
    while rc in (0, 301, 302):
        try:
            if rc != 0:
                conn = create_conn(address)[0]
            conn.request("GET", address, body=None, headers=headers)
        except SystemExit as e:
            raise
        except Exception as e:
            return None, None, f"Server request failed: {e}"
        response = conn.getresponse()
        rc = response.status

        # 301 means that the page address is wrong.
        if rc in (301, 302):
            # This response reading is ignored on purpose.
            _ = response.read()
            for x in str(response.msg).split("\n"):
                parts = x.split(": ", 1)
                if parts[0] == "Location":
                    if rc == 301:
                        sys.stderr.write(
                            f"{colorize('BAD', _('Location has moved: '))}{parts[1]}\n"
                        )
                    if rc == 302:
                        sys.stderr.write(
                            f"{colorize('BAD', _('Location has temporarily moved: '))}{parts[1]}\n"
                        )
                    address = parts[1]
                    break

    if rc not in (200, 206):
        return (
            None,
            rc,
            f"Server did not respond successfully ({response.status}: {response.reason})",
        )

    if dest:
        dest.write(response.read())
        return "", 0, ""

    return response.read(), 0, ""


def match_in_array(array, prefix="", suffix="", match_both=1, allow_overlap=0):

    warnings.warn(
        "portage.getbinpkg.match_in_array() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    myarray = []

    if not (prefix and suffix):
        match_both = 0

    for x in array:
        add_p = 0
        x_size = len(x)
        prefix_size = len(prefix)
        if prefix and x_size >= prefix_size and x[:prefix_size] == prefix:
            add_p = 1

        if match_both:
            if prefix and not add_p:  # Require both, but don't have first one.
                continue
        elif add_p:  # Only need one, and we have it.
            myarray.append(x[:])
            continue

        suffix_size = len(suffix)
        if not allow_overlap:  # Not allow to overlap prefix and suffix
            if x_size >= (prefix_size + suffix_size):
                pass
            else:
                continue  # Too short to match.
        else:
            pass  # Do whatever... We're overlapping.

        if suffix and x_size >= suffix_size and x[-len(suffix) :] == suffix:
            myarray.append(x)  # It matches
        else:
            continue  # Doesn't match.

    return myarray


def dir_get_list(baseurl, conn=None):
    """Takes a base url to connect to and read from.
    URI should be in the form <proto>://<site>[:port]<path>
    Connection is used for persistent connection instances."""

    warnings.warn(
        "portage.getbinpkg.dir_get_list() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    keepconnection = 0
    if conn:
        keepconnection = 1

    conn, protocol, address, params, headers = create_conn(baseurl, conn)

    listing = None
    if protocol in ["http", "https"]:
        if not address.endswith("/"):
            # http servers can return a 400 error here
            # if the address doesn't end with a slash.
            address = f"{address}/"
        page, rc, msg = make_http_request(conn, address, params, headers)

        if page:
            parser = ParseLinks()
            parser.feed(_unicode_decode(page))
            del page
            listing = parser.get_anchors()
        else:
            import portage.exception

            raise portage.exception.PortageException(
                _("Unable to get listing: %s %s") % (rc, msg)
            )
    elif protocol == "ftp":
        if address[-1] == "/":
            olddir = conn.pwd()
            conn.cwd(address)
            listing = conn.nlst()
            conn.cwd(olddir)
            del olddir
        else:
            listing = conn.nlst(address)
    elif protocol == "sftp":
        listing = conn.listdir(address)
    else:
        raise TypeError(_("Unknown protocol. '%s'") % protocol)

    if not keepconnection:
        conn.close()

    return listing


def file_get_metadata(baseurl, conn=None, chunk_size=3000):
    """Takes a base url to connect to and read from.
    URI should be in the form <proto>://<site>[:port]<path>
    Connection is used for persistent connection instances."""

    warnings.warn(
        "portage.getbinpkg.file_get_metadata() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    keepconnection = 1
    if conn:
        keepconnection = 0

    conn, protocol, address, params, headers = create_conn(baseurl, conn)

    if protocol in ["http", "https"]:
        headers["Range"] = f"bytes=-{chunk_size}"
        data, _x, _x = make_http_request(conn, address, params, headers)
    elif protocol in ["ftp"]:
        data, _x, _x = make_ftp_request(conn, address, -chunk_size)
    elif protocol == "sftp":
        f = conn.open(address)
        try:
            f.seek(-chunk_size, 2)
            data = f.read()
        finally:
            f.close()
    else:
        raise TypeError(_(f"Unknown protocol. '{protocol}'"))

    if data:
        xpaksize = portage.xpak.decodeint(data[-8:-4])
        if (xpaksize + 8) > chunk_size:
            myid = file_get_metadata(baseurl, conn, xpaksize + 8)
            if not keepconnection:
                conn.close()
            return myid
        xpak_data = data[len(data) - (xpaksize + 8) : -8]
        del data

        myid = portage.xpak.xsplit_mem(xpak_data)
        if not myid:
            myid = None, None
        del xpak_data
    else:
        myid = None, None

    if not keepconnection:
        conn.close()

    return myid


def file_get(
    baseurl=None, dest=None, conn=None, fcmd=None, filename=None, fcmd_vars=None
):
    """Takes a base url to connect to and read from.
    URI should be in the form <proto>://[user[:pass]@]<site>[:port]<path>"""

    if not fcmd:

        warnings.warn(
            "Use of portage.getbinpkg.file_get() without the fcmd "
            "parameter is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )

        return file_get_lib(baseurl, dest, conn)

    variables = {}

    if fcmd_vars is not None:
        variables.update(fcmd_vars)

    if "DISTDIR" not in variables:
        if dest is None:
            raise portage.exception.MissingParameter(
                _("fcmd_vars is missing required 'DISTDIR' key")
            )
        variables["DISTDIR"] = dest

    if "URI" not in variables:
        if baseurl is None:
            raise portage.exception.MissingParameter(
                _("fcmd_vars is missing required 'URI' key")
            )
        variables["URI"] = baseurl

    if "FILE" not in variables:
        if not filename:
            filename = os.path.basename(variables["URI"])
        variables["FILE"] = filename

    from portage.util import varexpand
    from portage.process import spawn

    myfetch = (varexpand(x, mydict=variables) for x in portage.util.shlex_split(fcmd))
    fd_pipes = {
        0: portage._get_stdin().fileno(),
        1: sys.__stdout__.fileno(),
        2: sys.__stdout__.fileno(),
    }
    sys.__stdout__.flush()
    sys.__stderr__.flush()
    retval = spawn(myfetch, env=os.environ.copy(), fd_pipes=fd_pipes)
    if retval != os.EX_OK:
        sys.stderr.write(_("Fetcher exited with a failure condition.\n"))
        return 0
    return 1


def file_get_lib(baseurl, dest, conn=None):
    """Takes a base url to connect to and read from.
    URI should be in the form <proto>://<site>[:port]<path>
    Connection is used for persistent connection instances."""

    warnings.warn(
        "portage.getbinpkg.file_get_lib() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    keepconnection = 0
    if conn:
        keepconnection = 1

    conn, protocol, address, params, headers = create_conn(baseurl, conn)

    sys.stderr.write(f"Fetching '{os.path.basename(address)}'\n")
    if protocol in ["http", "https"]:
        data, rc, _msg = make_http_request(conn, address, params, headers, dest=dest)
    elif protocol in ["ftp"]:
        data, rc, _msg = make_ftp_request(conn, address, dest=dest)
    elif protocol == "sftp":
        rc = 0
        try:
            f = conn.open(address)
        except SystemExit:
            raise
        except Exception:
            rc = 1
        else:
            try:
                if dest:
                    bufsize = 8192
                    while True:
                        data = f.read(bufsize)
                        if not data:
                            break
                        dest.write(data)
            finally:
                f.close()
    else:
        raise TypeError(_("Unknown protocol. '%s'") % protocol)

    if not keepconnection:
        conn.close()

    return rc


def dir_get_metadata(
    baseurl, conn=None, chunk_size=3000, verbose=1, usingcache=1, makepickle=None
):
    warnings.warn(
        "portage.getbinpkg.dir_get_metadata() is deprecated",
        DeprecationWarning,
        stacklevel=2,
    )

    keepconnection = 1
    if conn:
        keepconnection = 0

    cache_path = "/var/cache/edb"
    metadatafilename = os.path.join(cache_path, "remote_metadata.pickle")

    if not makepickle:
        makepickle = "/var/cache/edb/metadata.idx.most_recent"

    try:
        conn = create_conn(baseurl, conn)[0]
    except _all_errors as e:
        # ftplib.FTP(host) can raise errors like this:
        #   socket.error: (111, 'Connection refused')
        sys.stderr.write("!!! %s\n" % (e,))
        return {}

    out = sys.stdout
    try:
        metadatafile = open(
            _unicode_encode(
                metadatafilename, encoding=_encodings["fs"], errors="strict"
            ),
            "rb",
        )
        mypickle = pickle.Unpickler(metadatafile)
        try:
            mypickle.find_global = None
        except AttributeError:
            # TODO: If py3k, override Unpickler.find_class().
            pass
        metadata = mypickle.load()
        out.write(_("Loaded metadata pickle.\n"))
        out.flush()
        metadatafile.close()
    except (SystemExit, KeyboardInterrupt):
        raise
    except Exception:
        metadata = {}
    if baseurl not in metadata:
        metadata[baseurl] = {}
    if "indexname" not in metadata[baseurl]:
        metadata[baseurl]["indexname"] = ""
    if "timestamp" not in metadata[baseurl]:
        metadata[baseurl]["timestamp"] = 0
    if "unmodified" not in metadata[baseurl]:
        metadata[baseurl]["unmodified"] = 0
    if "data" not in metadata[baseurl]:
        metadata[baseurl]["data"] = {}

    if not os.access(cache_path, os.W_OK):
        sys.stderr.write(_("!!! Unable to write binary metadata to disk!\n"))
        sys.stderr.write(_(f"!!! Permission denied: '{cache_path}'\n"))
        return metadata[baseurl]["data"]

    import portage.exception

    try:
        filelist = dir_get_list(baseurl, conn)
    except portage.exception.PortageException as e:
        sys.stderr.write(_(f"!!! Error connecting to '{_hide_url_passwd(baseurl)}'.\n"))
        sys.stderr.write(f"!!! {e}\n")
        del e
        return metadata[baseurl]["data"]
    tbz2list = match_in_array(filelist, suffix=".tbz2")
    metalist = match_in_array(filelist, prefix="metadata.idx")
    del filelist

    # Determine if our metadata file is current.
    metalist.sort()
    metalist.reverse()  # makes the order new-to-old.
    for mfile in metalist:
        if usingcache and (
            (metadata[baseurl]["indexname"] != mfile)
            or (metadata[baseurl]["timestamp"] < int(time.time() - (60 * 60 * 24)))
        ):
            # Try to download new cache until we succeed on one.
            data = ""
            for trynum in [1, 2, 3]:
                mytempfile = tempfile.TemporaryFile()
                try:
                    file_get(baseurl + "/" + mfile, mytempfile, conn)
                    if mytempfile.tell() > len(data):
                        mytempfile.seek(0)
                        data = mytempfile.read()
                except ValueError as e:
                    sys.stderr.write("--- %s\n" % str(e))
                    if trynum < 3:
                        sys.stderr.write(_("Retrying...\n"))
                    sys.stderr.flush()
                    mytempfile.close()
                    continue
                if match_in_array([mfile], suffix=".gz"):
                    out.write("gzip'd\n")
                    out.flush()
                    try:
                        import gzip

                        mytempfile.seek(0)
                        gzindex = gzip.GzipFile(mfile[:-3], "rb", 9, mytempfile)
                        data = gzindex.read()
                    except SystemExit as e:
                        raise
                    except Exception as e:
                        mytempfile.close()
                        sys.stderr.write(_("!!! Failed to use gzip: ") + str(e) + "\n")
                        sys.stderr.flush()
                    mytempfile.close()
                try:
                    metadata[baseurl]["data"] = pickle.loads(data)
                    del data
                    metadata[baseurl]["indexname"] = mfile
                    metadata[baseurl]["timestamp"] = int(time.time())
                    metadata[baseurl]["modified"] = 0  # It's not, right after download.
                    out.write(_("Pickle loaded.\n"))
                    out.flush()
                    break
                except SystemExit as e:
                    raise
                except Exception as e:
                    sys.stderr.write(f"!!! Failed to read data from index: {mfile}\n")
                    sys.stderr.write(f"!!! {e}")
                    sys.stderr.flush()
            try:
                metadatafile = open(
                    _unicode_encode(
                        metadatafilename, encoding=_encodings["fs"], errors="strict"
                    ),
                    "wb",
                )
                pickle.dump(metadata, metadatafile, protocol=2)
                metadatafile.close()
            except SystemExit as e:
                raise
            except Exception as e:
                sys.stderr.write(_("!!! Failed to write binary metadata to disk!\n"))
                sys.stderr.write(f"!!! {e}\n")
                sys.stderr.flush()
            break
    # We may have metadata... now we run through the tbz2 list and check.

    class CacheStats:
        from time import time

        def __init__(self, out):
            self.misses = 0
            self.hits = 0
            self.last_update = 0
            self.out = out
            self.min_display_latency = 0.2

        def update(self):
            cur_time = self.time()
            if cur_time - self.last_update >= self.min_display_latency:
                self.last_update = cur_time
                self.display()

        def display(self):
            self.out.write(
                "".join(
                    (
                        "\r",
                        colorize("WARN", _(f"cache miss: '{self.misses}'")),
                        " --- ",
                        colorize("GOOD", _(f"cache hit: '{self.hits}'")),
                    )
                )
            )
            self.out.flush()

    cache_stats = CacheStats(out)
    have_tty = os.environ.get("TERM") != "dumb" and out.isatty()
    if have_tty:
        cache_stats.display()
    binpkg_filenames = set()
    for x in tbz2list:
        x = os.path.basename(x)
        binpkg_filenames.add(x)
        if x not in metadata[baseurl]["data"]:
            cache_stats.misses += 1
            if have_tty:
                cache_stats.update()
            metadata[baseurl]["modified"] = 1
            myid = None
            for _x in range(3):
                try:
                    myid = file_get_metadata(
                        "/".join((baseurl.rstrip("/"), x.lstrip("/"))), conn, chunk_size
                    )
                    break
                except http_client_BadStatusLine:
                    # Sometimes this error is thrown from conn.getresponse() in
                    # make_http_request().  The docstring for this error in
                    # httplib.py says "Presumably, the server closed the
                    # connection before sending a valid response".
                    conn = create_conn(baseurl)[0]
                except http_client_ResponseNotReady:
                    # With some http servers this error is known to be thrown
                    # from conn.getresponse() in make_http_request() when the
                    # remote file does not have appropriate read permissions.
                    # Maybe it's possible to recover from this exception in
                    # cases though, so retry.
                    conn = create_conn(baseurl)[0]

            if myid and myid[0]:
                metadata[baseurl]["data"][x] = make_metadata_dict(myid)
            elif verbose:
                sys.stderr.write(
                    f"{colorize('BAD', _('!!! Failed to retrieve metadata on: '))}{x}\n"
                )
                sys.stderr.flush()
        else:
            cache_stats.hits += 1
            if have_tty:
                cache_stats.update()
    cache_stats.display()
    # Cleanse stale cache for files that don't exist on the server anymore.
    stale_cache = set(metadata[baseurl]["data"]).difference(binpkg_filenames)
    if stale_cache:
        for x in stale_cache:
            del metadata[baseurl]["data"][x]
        metadata[baseurl]["modified"] = 1
    del stale_cache
    del binpkg_filenames
    out.write("\n")
    out.flush()

    try:
        if "modified" in metadata[baseurl] and metadata[baseurl]["modified"]:
            metadata[baseurl]["timestamp"] = int(time.time())
            metadatafile = open(
                _unicode_encode(
                    metadatafilename, encoding=_encodings["fs"], errors="strict"
                ),
                "wb",
            )
            pickle.dump(metadata, metadatafile, protocol=2)
            metadatafile.close()
        if makepickle:
            metadatafile = open(
                _unicode_encode(makepickle, encoding=_encodings["fs"], errors="strict"),
                "wb",
            )
            pickle.dump(metadata[baseurl]["data"], metadatafile, protocol=2)
            metadatafile.close()
    except SystemExit as e:
        raise
    except Exception as e:
        sys.stderr.write(_("!!! Failed to write binary metadata to disk!\n"))
        sys.stderr.write("!!! " + str(e) + "\n")
        sys.stderr.flush()

    if not keepconnection:
        conn.close()

    return metadata[baseurl]["data"]


def _cmp_cpv(d1, d2):
    cpv1 = d1["CPV"]
    cpv2 = d2["CPV"]
    if cpv1 > cpv2:
        return 1
    if cpv1 == cpv2:
        return 0
    return -1


class PackageIndex:
    def __init__(
        self,
        allowed_pkg_keys=None,
        default_header_data=None,
        default_pkg_data=None,
        inherited_keys=None,
        translated_keys=None,
    ):

        self._pkg_slot_dict = None
        if allowed_pkg_keys:
            self._pkg_slot_dict = slot_dict_class(allowed_pkg_keys)

        self._default_header_data = default_header_data
        self._default_pkg_data = default_pkg_data
        self._inherited_keys = inherited_keys
        self._write_translation_map = {}
        self._read_translation_map = {}
        if translated_keys:
            self._write_translation_map.update(translated_keys)
            self._read_translation_map.update(((y, x) for (x, y) in translated_keys))
        self.header = {}
        if self._default_header_data:
            self.header.update(self._default_header_data)
        self.packages = []
        self.modified = True

    def _readpkgindex(self, pkgfile, pkg_entry=True):
        d = {}
        allowed_keys = None
        if self._pkg_slot_dict and pkg_entry:
            d = self._pkg_slot_dict()
            allowed_keys = d.allowed_keys

        for line in pkgfile:
            line = line.rstrip("\n")
            if not line:
                break
            line = line.split(":", 1)
            if not len(line) == 2:
                continue
            k, v = line
            if v:
                v = v[1:]
            k = self._read_translation_map.get(k, k)
            if allowed_keys is not None and k not in allowed_keys:
                continue
            d[k] = v
        return d

    def _writepkgindex(self, pkgfile, items):
        for k, v in items:
            pkgfile.write(f"{self._write_translation_map.get(k, k)}: {v}\n")
        pkgfile.write("\n")

    def read(self, pkgfile):
        self.readHeader(pkgfile)
        self.readBody(pkgfile)

    def readHeader(self, pkgfile):
        self.header.update(self._readpkgindex(pkgfile, pkg_entry=False))

    def readBody(self, pkgfile):
        while True:
            d = self._readpkgindex(pkgfile)
            if not d:
                break
            mycpv = d.get("CPV")
            if not mycpv:
                continue
            if self._default_pkg_data:
                for k, v in self._default_pkg_data.items():
                    d.setdefault(k, v)
            if self._inherited_keys:
                for k in self._inherited_keys:
                    v = self.header.get(k)
                    if v:
                        d.setdefault(k, v)
            self.packages.append(d)

    def write(self, pkgfile):
        if self.modified:
            self.header["TIMESTAMP"] = str(int(time.time()))
            self.header["PACKAGES"] = str(len(self.packages))
        keys = list(self.header)
        keys.sort()
        self._writepkgindex(
            pkgfile, [(k, self.header[k]) for k in keys if self.header[k]]
        )
        for metadata in sorted(self.packages, key=portage.util.cmp_sort_key(_cmp_cpv)):
            metadata = metadata.copy()
            if self._inherited_keys:
                for k in self._inherited_keys:
                    v = self.header.get(k)
                    if v and v == metadata.get(k):
                        del metadata[k]
            if self._default_pkg_data:
                for k, v in self._default_pkg_data.items():
                    if metadata.get(k) == v:
                        metadata.pop(k, None)
            keys = list(metadata)
            keys.sort()
            self._writepkgindex(
                pkgfile, ((k, metadata[k]) for k in keys if metadata[k])
            )



# Copright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import io
import random
import tarfile
import tempfile
from functools import partial
from os import urandom

from portage.gpkg import gpkg
from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.exception import InvalidSignature
from portage.gpg import GPG


class test_gpkg_metadata_url_case(TestCase):
    def httpd(self, directory, port):
        try:
            import http.server
            import socketserver
        except ImportError:
            self.skipTest("http server not exits")

        Handler = partial(http.server.SimpleHTTPRequestHandler, directory=directory)

        with socketserver.TCPServer(("127.0.0.1", port), Handler) as httpd:
            httpd.serve_forever()

    def start_http_server(self, directory, port):
        try:
            import threading
        except ImportError:
            self.skipTest("threading module not exists")

        server = threading.Thread(
            target=self.httpd, args=(directory, port), daemon=True
        )
        server.start()
        return server

    def test_gpkg_get_metadata_url(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'BINPKG_COMPRESS="gzip"',
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()
        try:
            settings = playground.settings
            for _ in range(0, 5):
                port = random.randint(30000, 60000)
                try:
                    server = self.start_http_server(tmpdir, port)
                except OSError:
                    continue
                break

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            meta = {
                "test1": b"{abcdefghijklmnopqrstuvwxyz, 1234567890}",
                "test2": urandom(102400),
            }

            test_gpkg.compress(os.path.join(tmpdir, "orig"), meta)

            meta_from_url = test_gpkg.get_metadata_url(
                "http://127.0.0.1:" + str(port) + "/test.gpkg.tar"
            )

            self.assertEqual(meta, meta_from_url)
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_get_metadata_url_unknown_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'BINPKG_COMPRESS="gzip"',
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()
        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()

            for _ in range(0, 5):
                port = random.randint(30000, 60000)
                try:
                    server = self.start_http_server(tmpdir, port)
                except OSError:
                    continue
                break

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test-1.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            meta = {
                "test1": b"{abcdefghijklmnopqrstuvwxyz, 1234567890}",
                "test2": urandom(102400),
            }

            test_gpkg.compress(os.path.join(tmpdir, "orig"), meta)

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == "test/metadata.tar.gz":
                            sig = b"""
-----BEGIN PGP SIGNATURE-----

iHUEABYIAB0WIQRVhCbPGi/rhGTq4nV+k2dcK9uyIgUCXw4ehAAKCRB+k2dcK9uy
IkCfAP49AOYjzuQPP0n5P0SGCINnAVEXN7QLQ4PurY/lt7cT2gEAq01stXjFhrz5
87Koh+ND2r5XfQsz3XeBqbb/BpmbEgo=
=sc5K
-----END PGP SIGNATURE-----
"""
                            data = io.BytesIO(sig)
                            f.size = len(sig)
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            test_gpkg = gpkg(settings, "test")
            self.assertRaises(
                InvalidSignature,
                test_gpkg.get_metadata_url,
                "http://127.0.0.1:" + str(port) + "/test-2.gpkg.tar",
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["deprecated_profile_check"]

import io

import portage
from portage import os, _encodings, _unicode_encode
from portage.const import DEPRECATED_PROFILE_FILE
from portage.localization import _
from portage.output import colorize
from portage.util import writemsg


def deprecated_profile_check(settings=None):
    config_root = None
    eprefix = None
    deprecated_profile_file = None
    if settings is not None:
        config_root = settings["PORTAGE_CONFIGROOT"]
        eprefix = settings["EPREFIX"]
        for x in reversed(settings._locations_manager.profiles_complex):
            if x.show_deprecated_warning:
                deprecated_profile_file = os.path.join(x.location, "deprecated")
                if os.access(deprecated_profile_file, os.R_OK):
                    break
        else:
            deprecated_profile_file = None

    if deprecated_profile_file is None:
        deprecated_profile_file = os.path.join(
            config_root or "/", DEPRECATED_PROFILE_FILE
        )
        if not os.access(deprecated_profile_file, os.R_OK):
            deprecated_profile_file = os.path.join(
                config_root or "/", "etc", "make.profile", "deprecated"
            )
            if not os.access(deprecated_profile_file, os.R_OK):
                return

    with io.open(
        _unicode_encode(
            deprecated_profile_file, encoding=_encodings["fs"], errors="strict"
        ),
        mode="r",
        encoding=_encodings["content"],
        errors="replace",
    ) as f:
        dcontent = f.readlines()
    writemsg(
        colorize(
            "BAD",
            _("\n!!! Your current profile is " "deprecated and not supported anymore."),
        )
        + "\n",
        noiselevel=-1,
    )
    writemsg(
        colorize("BAD", _("!!! Use eselect profile to update your " "profile.")) + "\n",
        noiselevel=-1,
    )
    if not dcontent:
        writemsg(
            colorize("BAD", _("!!! Please refer to the " "Gentoo Upgrading Guide."))
            + "\n",
            noiselevel=-1,
        )
        return True
    newprofile = dcontent[0].rstrip("\n")
    writemsg(
        colorize(
            "BAD", _("!!! Please upgrade to the " "following profile if possible:")
        )
        + "\n\n",
        noiselevel=-1,
    )
    writemsg(8 * " " + colorize("GOOD", newprofile) + "\n\n", noiselevel=-1)
    if len(dcontent) > 1:
        writemsg(_("To upgrade do the following steps:\n"), noiselevel=-1)
        for myline in dcontent[1:]:
            writemsg(myline, noiselevel=-1)
        writemsg("\n\n", noiselevel=-1)
    else:
        writemsg(_("You may use the following command to upgrade:\n\n"), noiselevel=-1)
        writemsg(
            8 * " " + colorize("INFORM", "eselect profile set " + newprofile) + "\n\n",
            noiselevel=-1,
        )

    if settings is not None:
        main_repo_loc = settings.repositories.mainRepoLocation()
        new_profile_path = os.path.join(
            main_repo_loc, "profiles", newprofile.rstrip("\n")
        )

        if os.path.isdir(new_profile_path):
            new_config = portage.config(
                config_root=config_root,
                config_profile_path=new_profile_path,
                eprefix=eprefix,
            )

            if not new_config.profiles:
                writemsg(
                    "\n %s %s\n"
                    % (
                        colorize("WARN", "*"),
                        _(
                            "You must update portage before you "
                            "can migrate to the above profile."
                        ),
                    ),
                    noiselevel=-1,
                )
                writemsg(
                    " %s %s\n\n"
                    % (
                        colorize("WARN", "*"),
                        _(
                            "In order to update portage, "
                            "run 'emerge --oneshot sys-apps/portage'."
                        ),
                    ),
                    noiselevel=-1,
                )

    return True



# data.py -- Calculated/Discovered Data Values
# Copyright 1998-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import grp
import os
import platform
import pwd

import portage
from portage.localization import _

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.output:colorize",
    "portage.util:writemsg",
    "portage.util.path:first_existing",
    "subprocess",
)

ostype = platform.system()
userland = "GNU"
if ostype == "DragonFly" or ostype.endswith("BSD"):
    userland = "BSD"

lchown = getattr(os, "lchown", None)

if not lchown:
    if ostype == "Darwin":

        def lchown(*_args, **_kwargs):
            pass

    else:

        def lchown(*_args, **_kwargs):
            writemsg(
                colorize("BAD", "!!!")
                + _(
                    " It seems that os.lchown does not"
                    " exist.  Please rebuild python.\n"
                ),
                noiselevel=-1,
            )

        lchown()

lchown = portage._unicode_func_wrapper(lchown)


def _target_eprefix():
    """
    Calculate the target EPREFIX, which may be different from
    portage.const.EPREFIX due to cross-prefix support. The result
    is equivalent to portage.settings["EPREFIX"], but the calculation
    is done without the expense of instantiating portage.settings.
    @rtype: str
    @return: the target EPREFIX
    """
    eprefix = os.environ.get("EPREFIX", portage.const.EPREFIX)
    if eprefix:
        eprefix = portage.util.normalize_path(eprefix)
    return eprefix


def _target_root():
    """
    Calculate the target ROOT. The result is equivalent to
    portage.settings["ROOT"], but the calculation
    is done without the expense of instantiating portage.settings.
    @rtype: str
    @return: the target ROOT (always ends with a slash)
    """
    root = os.environ.get("ROOT")
    if not root:
        # Handle either empty or unset ROOT.
        root = os.sep
    root = portage.util.normalize_path(root)
    return f"{root.rstrip(os.sep)}{os.sep}"


def portage_group_warning():
    warn_prefix = colorize("BAD", "*** WARNING ***  ")
    mylines = (
        "For security reasons, only system administrators should be",
        "allowed in the portage group.  Untrusted users or processes",
        "can potentially exploit the portage group for attacks such as",
        "local privilege escalation.",
    )
    for x in mylines:
        writemsg(warn_prefix, noiselevel=-1)
        writemsg(x, noiselevel=-1)
        writemsg("\n", noiselevel=-1)
    writemsg("\n", noiselevel=-1)


# Portage has 3 security levels that depend on the uid and gid of the main
# process and are assigned according to the following table:
#
# Privileges  secpass  uid    gid
# normal      0        any    any
# group       1        any    portage_gid
# super       2        0      any
#
# If the "wheel" group does not exist then wheelgid falls back to 0.
# If the "portage" group does not exist then portage_uid falls back to wheelgid.

# If the current user is not root, but has write access to the
# EROOT directory (not due to the 0002 bit), then use "unprivileged"
# mode which sets secpass = 2 and uses the UID and GID of the EROOT
# directory to generate default PORTAGE_INST_GID, PORTAGE_INST_UID,
# PORTAGE_USERNAME, and PORTAGE_GRPNAME settings.


def _unprivileged_mode(eroot, eroot_st):
    return (
        os.getuid() != 0 and os.access(eroot, os.W_OK) and not eroot_st.st_mode & 0o0002
    )


uid = os.getuid()
wheelgid = 0
try:
    wheelgid = grp.getgrnam("wheel")[2]
except KeyError:
    pass

# The portage_uid and portage_gid global constants, and others that
# depend on them are initialized lazily, in order to allow configuration
# via make.conf. Eventually, these constants may be deprecated in favor
# of config attributes, since it's conceivable that multiple
# configurations with different constants could be used simultaneously.
_initialized_globals = set()


def _get_global(k):
    if k in _initialized_globals:
        return globals()[k]

    if k == "secpass":
        unprivileged = False
        if hasattr(portage, "settings"):
            unprivileged = "unprivileged" in portage.settings.features
        else:
            # The config class has equivalent code, but we also need to
            # do it here if _disable_legacy_globals() has been called.
            eroot_or_parent = first_existing(
                os.path.join(_target_root(), _target_eprefix().lstrip(os.sep))
            )
            try:
                eroot_st = os.stat(eroot_or_parent)
            except OSError:
                pass
            else:
                unprivileged = _unprivileged_mode(eroot_or_parent, eroot_st)

        v = 0
        if uid == 0 or unprivileged:
            v = 2
        elif _get_global("portage_gid") in os.getgroups():
            v = 1

    elif k in ("portage_gid", "portage_uid"):
        # Discover the uid and gid of the portage user/group
        keyerror = False
        try:
            portage_uid = pwd.getpwnam(_get_global("_portage_username")).pw_uid
        except KeyError:
            keyerror = True
            portage_uid = 0

        try:
            portage_gid = grp.getgrnam(_get_global("_portage_grpname")).gr_gid
        except KeyError:
            keyerror = True
            portage_gid = 0

        # Suppress this error message if both PORTAGE_GRPNAME and
        # PORTAGE_USERNAME are set to "root", for things like
        # Android (see bug #454060).
        if keyerror and not (
            _get_global("_portage_username") == "root"
            and _get_global("_portage_grpname") == "root"
        ):
            writemsg(
                colorize("BAD", _("portage: 'portage' user or group missing.")) + "\n",
                noiselevel=-1,
            )
            writemsg(
                _(
                    "         For the defaults, line 1 goes into passwd, "
                    "and 2 into group.\n"
                ),
                noiselevel=-1,
            )
            writemsg(
                colorize(
                    "GOOD",
                    "         portage:x:250:250:portage:/var/tmp/portage:/bin/false",
                )
                + "\n",
                noiselevel=-1,
            )
            writemsg(
                colorize("GOOD", "         portage::250:portage") + "\n", noiselevel=-1
            )
            portage_group_warning()

        globals()["portage_gid"] = portage_gid
        _initialized_globals.add("portage_gid")
        globals()["portage_uid"] = portage_uid
        _initialized_globals.add("portage_uid")

        if k == "portage_gid":
            return portage_gid
        if k == "portage_uid":
            return portage_uid
        raise AssertionError("unknown name: %s" % k)

    elif k == "userpriv_groups":
        v = [_get_global("portage_gid")]
        if secpass >= 2:
            # Get a list of group IDs for the portage user. Do not use
            # grp.getgrall() since it is known to trigger spurious
            # SIGPIPE problems with nss_ldap.
            encoding = portage._encodings["content"]
            cmd = (
                portage._unicode_encode(x, encoding=encoding, errors="strict")
                for x in ("id", "-G", _portage_username)
            )
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            myoutput = proc.communicate()[0]
            status = proc.wait()
            if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:

                def check(x):
                    try:
                        return int(x)
                    except ValueError:
                        return None

                unicode_decode = portage._unicode_decode(
                    myoutput, encoding=encoding, errors="strict"
                )
                checked_v = (check(x) for x in unicode_decode.split())
                filtered_v = (x for x in checked_v if x)
                v = sorted(set(filtered_v))

    # Avoid instantiating portage.settings when the desired
    # variable is set in os.environ.
    elif k in ("_portage_grpname", "_portage_username"):
        v = "portage"
        env_key = "PORTAGE_USERNAME"
        if k == "_portage_grpname":
            env_key = "PORTAGE_GRPNAME"

        if env_key in os.environ:
            v = os.environ[env_key]
        elif hasattr(portage, "settings"):
            v = portage.settings.get(env_key)
        else:
            # The config class has equivalent code, but we also need to
            # do it here if _disable_legacy_globals() has been called.
            eroot_or_parent = first_existing(
                os.path.join(_target_root(), _target_eprefix().lstrip(os.sep))
            )
            try:
                eroot_st = os.stat(eroot_or_parent)
            except OSError:
                pass
            else:
                if _unprivileged_mode(eroot_or_parent, eroot_st):
                    if k == "_portage_grpname":
                        try:
                            grp_struct = grp.getgrgid(eroot_st.st_gid)
                        except KeyError:
                            pass
                        else:
                            v = grp_struct.gr_name
                    else:
                        try:
                            pwd_struct = pwd.getpwuid(eroot_st.st_uid)
                        except KeyError:
                            pass
                        else:
                            v = pwd_struct.pw_name
    else:
        raise AssertionError("unknown name: %s" % k)

    globals()[k] = v
    _initialized_globals.add(k)
    return v


class _GlobalProxy(portage.proxy.objectproxy.ObjectProxy):
    __slots__ = ("_name",)

    def __init__(self, name):
        portage.proxy.objectproxy.ObjectProxy.__init__(self)
        object.__setattr__(self, "_name", name)

    def _get_target(self):
        return _get_global(object.__getattribute__(self, "_name"))


for k in (
    "portage_gid",
    "portage_uid",
    "secpass",
    "userpriv_groups",
    "_portage_grpname",
    "_portage_username",
):
    globals()[k] = _GlobalProxy(k)
del k


def _init(settings):
    """
    Use config variables like PORTAGE_GRPNAME and PORTAGE_USERNAME to
    initialize global variables. This allows settings to come from make.conf
    instead of requiring them to be set in the calling environment.
    """
    if (
        "_portage_grpname" not in _initialized_globals
        and "_portage_username" not in _initialized_globals
    ):

        # Prevents "TypeError: expected string" errors
        # from grp.getgrnam() with PyPy
        native_string = platform.python_implementation() == "PyPy"

        v = settings.get("PORTAGE_GRPNAME", "portage")
        if native_string:
            v = portage._native_string(v)
        globals()["_portage_grpname"] = v
        _initialized_globals.add("_portage_grpname")

        v = settings.get("PORTAGE_USERNAME", "portage")
        if native_string:
            v = portage._native_string(v)
        globals()["_portage_username"] = v
        _initialized_globals.add("_portage_username")

    if "secpass" not in _initialized_globals:
        v = 0
        if uid == 0 or "unprivileged" in settings.features:
            v = 2
        elif portage_gid in os.getgroups():
            v = 1
        globals()["secpass"] = v
        _initialized_globals.add("secpass")



# Copyright 2015-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["get_cpu_count"]

# Before you set out to change this function, figure out what you're really
# asking:
#
# - How many CPUs exist in this system (e.g. that the kernel is aware of?)
#   This is 'getconf _NPROCESSORS_CONF' / get_nprocs_conf(3)
#   In modern Linux, implemented by counting CPUs in /sys/devices/system/cpu/
#
# - How many CPUs in this system are ONLINE right now?
#   This is 'getconf _NPROCESSORS_ONLN' / get_nprocs(3)
#   In modern Linux, implemented by parsing /sys/devices/system/cpu/online
#
# - How many CPUs are available to this program?
#   This is 'nproc' / sched_getaffinity(2), which is implemented in modern
#   Linux kernels by querying the kernel scheduler; This might not be available
#   in some non-Linux systems!
#
# - How many CPUs are available to this thread?
#   This is pthread_getaffinity_np(3)
#
# As a further warning, the results returned by this function can differ
# between runs, if altered by the scheduler or other external factors.


def get_cpu_count():
    """
    Try to obtain the number of CPUs available to this process.

    @return: Number of CPUs or None if unable to obtain.
    """
    try:
        import os

        # This was introduced in Python 3.3 only, but exists in Linux
        # all the way back to the 2.5.8 kernel.
        # This NOT available in FreeBSD!
        return len(os.sched_getaffinity(0))
    except (ImportError, NotImplementedError, AttributeError):
        pass

    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        return None



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import copy
import io
import sys

import portage
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage import _unicode_decode
from portage.checksum import _hash_filter
from portage.elog.messages import eerror
from portage.package.ebuild.fetch import (
    _check_distfile,
    _drop_privs_userfetch,
    _want_userfetch,
    fetch,
)
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.ForkProcess import ForkProcess
from portage.util._pty import _create_pty_or_pipe
from _emerge.CompositeTask import CompositeTask


class EbuildFetcher(CompositeTask):

    __slots__ = (
        "config_pool",
        "ebuild_path",
        "fetchonly",
        "fetchall",
        "logfile",
        "pkg",
        "prefetch",
        "_fetcher_proc",
    )

    def __init__(self, **kwargs):
        CompositeTask.__init__(self, **kwargs)
        self._fetcher_proc = _EbuildFetcherProcess(**kwargs)

    def async_already_fetched(self, settings):
        """
        Returns True if all files already exist locally and have correct
        digests, otherwise return False. When returning True, appropriate
        digest checking messages are produced for display and/or logging.
        When returning False, no messages are produced, since we assume
        that a fetcher process will later be executed in order to produce
        such messages. This will raise InvalidDependString if SRC_URI is
        invalid.
        """
        return self._fetcher_proc.async_already_fetched(settings)

    def _start(self):
        self._start_task(
            AsyncTaskFuture(future=self._fetcher_proc._async_uri_map()),
            self._start_fetch,
        )

    def _start_fetch(self, uri_map_task):
        self._assert_current(uri_map_task)
        if uri_map_task.cancelled:
            self._default_final_exit(uri_map_task)
            return

        try:
            uri_map = uri_map_task.future.result()
        except portage.exception.InvalidDependString as e:
            msg_lines = []
            msg = "Fetch failed for '%s' due to invalid SRC_URI: %s" % (self.pkg.cpv, e)
            msg_lines.append(msg)
            self._fetcher_proc._eerror(msg_lines)
            self._current_task = None
            self.returncode = 1
            self._async_wait()
            return

        # First get the SRC_URI metadata (it's not cached in self.pkg.metadata
        # because some packages have an extremely large SRC_URI value).
        self._start_task(
            AsyncTaskFuture(
                future=self.pkg.root_config.trees["porttree"].dbapi.async_aux_get(
                    self.pkg.cpv, ["SRC_URI"], myrepo=self.pkg.repo, loop=self.scheduler
                )
            ),
            self._start_with_metadata,
        )

    def _start_with_metadata(self, aux_get_task):
        self._assert_current(aux_get_task)
        if aux_get_task.cancelled:
            self._default_final_exit(aux_get_task)
            return

        (self._fetcher_proc.src_uri,) = aux_get_task.future.result()
        self._start_task(self._fetcher_proc, self._default_final_exit)


class _EbuildFetcherProcess(ForkProcess):

    __slots__ = (
        "config_pool",
        "ebuild_path",
        "fetchonly",
        "fetchall",
        "pkg",
        "prefetch",
        "src_uri",
        "_digests",
        "_manifest",
        "_settings",
        "_uri_map",
    )

    def async_already_fetched(self, settings):
        result = self.scheduler.create_future()

        def uri_map_done(uri_map_future):
            if uri_map_future.cancelled():
                result.cancel()
                return

            if uri_map_future.exception() is not None or result.cancelled():
                if not result.cancelled():
                    result.set_exception(uri_map_future.exception())
                return

            uri_map = uri_map_future.result()
            if uri_map:
                result.set_result(self._check_already_fetched(settings, uri_map))
            else:
                result.set_result(True)

        uri_map_future = self._async_uri_map()
        result.add_done_callback(
            lambda result: uri_map_future.cancel() if result.cancelled() else None
        )
        uri_map_future.add_done_callback(uri_map_done)
        return result

    def _check_already_fetched(self, settings, uri_map):
        digests = self._get_digests()
        distdir = settings["DISTDIR"]
        allow_missing = self._get_manifest().allow_missing

        for filename in uri_map:
            # Use stat rather than lstat since fetch() creates
            # symlinks when PORTAGE_RO_DISTDIRS is used.
            try:
                st = os.stat(os.path.join(distdir, filename))
            except OSError:
                return False
            if st.st_size == 0:
                return False
            expected_size = digests.get(filename, {}).get("size")
            if expected_size is None:
                continue
            if st.st_size != expected_size:
                return False

        hash_filter = _hash_filter(settings.get("PORTAGE_CHECKSUM_FILTER", ""))
        if hash_filter.transparent:
            hash_filter = None
        stdout_orig = sys.stdout
        stderr_orig = sys.stderr
        global_havecolor = portage.output.havecolor
        out = io.StringIO()
        eout = portage.output.EOutput()
        eout.quiet = settings.get("PORTAGE_QUIET") == "1"
        success = True
        try:
            sys.stdout = out
            sys.stderr = out
            if portage.output.havecolor:
                portage.output.havecolor = not self.background

            for filename in uri_map:
                mydigests = digests.get(filename)
                if mydigests is None:
                    if not allow_missing:
                        success = False
                        break
                    continue
                ok, st = _check_distfile(
                    os.path.join(distdir, filename),
                    mydigests,
                    eout,
                    show_errors=False,
                    hash_filter=hash_filter,
                )
                if not ok:
                    success = False
                    break
        except portage.exception.FileNotFound:
            # A file disappeared unexpectedly.
            return False
        finally:
            sys.stdout = stdout_orig
            sys.stderr = stderr_orig
            portage.output.havecolor = global_havecolor

        if success:
            # When returning unsuccessfully, no messages are produced, since
            # we assume that a fetcher process will later be executed in order
            # to produce such messages.
            msg = out.getvalue()
            if msg:
                self.scheduler.output(msg, log_path=self.logfile)

        return success

    def _start(self):

        root_config = self.pkg.root_config
        portdb = root_config.trees["porttree"].dbapi
        ebuild_path = self._get_ebuild_path()
        # This is initialized by an earlier _async_uri_map call.
        uri_map = self._uri_map

        if not uri_map:
            # Nothing to fetch.
            self.returncode = os.EX_OK
            self._async_wait()
            return

        settings = self.config_pool.allocate()
        settings.setcpv(self.pkg)
        settings.configdict["pkg"]["SRC_URI"] = self.src_uri
        portage.doebuild_environment(ebuild_path, "fetch", settings=settings, db=portdb)

        if self.prefetch and self._prefetch_size_ok(uri_map, settings, ebuild_path):
            self.config_pool.deallocate(settings)
            self.returncode = os.EX_OK
            self._async_wait()
            return

        nocolor = settings.get("NOCOLOR")

        if self.prefetch:
            settings["PORTAGE_PARALLEL_FETCHONLY"] = "1"

        if self.background:
            nocolor = "true"

        if nocolor is not None:
            settings["NOCOLOR"] = nocolor

        self._settings = settings
        self.log_filter_file = settings.get("PORTAGE_LOG_FILTER_FILE_CMD")
        ForkProcess._start(self)

        # Free settings now since it's no longer needed in
        # this process (the subprocess has a private copy).
        self.config_pool.deallocate(settings)
        settings = None
        self._settings = None

    def _run(self):
        # Force consistent color output, in case we are capturing fetch
        # output through a normal pipe due to unavailability of ptys.
        portage.output.havecolor = self._settings.get("NOCOLOR") not in ("yes", "true")

        # For userfetch, drop privileges for the entire fetch call, in
        # order to handle DISTDIR on NFS with root_squash for bug 601252.
        if _want_userfetch(self._settings):
            _drop_privs_userfetch(self._settings)

        rval = 1
        allow_missing = (
            self._get_manifest().allow_missing or "digest" in self._settings.features
        )
        if fetch(
            self._uri_map,
            self._settings,
            fetchonly=self.fetchonly,
            digests=copy.deepcopy(self._get_digests()),
            allow_missing_digests=allow_missing,
        ):
            rval = os.EX_OK
        return rval

    def _get_ebuild_path(self):
        if self.ebuild_path is not None:
            return self.ebuild_path
        portdb = self.pkg.root_config.trees["porttree"].dbapi
        self.ebuild_path = portdb.findname(self.pkg.cpv, myrepo=self.pkg.repo)
        if self.ebuild_path is None:
            raise AssertionError("ebuild not found for '%s'" % self.pkg.cpv)
        return self.ebuild_path

    def _get_manifest(self):
        if self._manifest is None:
            pkgdir = os.path.dirname(self._get_ebuild_path())
            self._manifest = (
                self.pkg.root_config.settings.repositories.get_repo_for_location(
                    os.path.dirname(os.path.dirname(pkgdir))
                ).load_manifest(pkgdir, None)
            )
        return self._manifest

    def _get_digests(self):
        if self._digests is None:
            self._digests = self._get_manifest().getTypeDigests("DIST")
        return self._digests

    def _async_uri_map(self):
        """
        This calls the portdbapi.async_fetch_map method and returns the
        resulting Future (may contain InvalidDependString exception).
        """
        if self._uri_map is not None:
            result = self.scheduler.create_future()
            result.set_result(self._uri_map)
            return result

        pkgdir = os.path.dirname(self._get_ebuild_path())
        mytree = os.path.dirname(os.path.dirname(pkgdir))
        use = None
        if not self.fetchall:
            use = self.pkg.use.enabled
        portdb = self.pkg.root_config.trees["porttree"].dbapi

        def cache_result(result):
            try:
                self._uri_map = result.result()
            except Exception:
                # The caller handles this when it retrieves the result.
                pass

        result = portdb.async_fetch_map(
            self.pkg.cpv, useflags=use, mytree=mytree, loop=self.scheduler
        )
        result.add_done_callback(cache_result)
        return result

    def _prefetch_size_ok(self, uri_map, settings, ebuild_path):
        distdir = settings["DISTDIR"]

        sizes = {}
        for filename in uri_map:
            # Use stat rather than lstat since portage.fetch() creates
            # symlinks when PORTAGE_RO_DISTDIRS is used.
            try:
                st = os.stat(os.path.join(distdir, filename))
            except OSError:
                return False
            if st.st_size == 0:
                return False
            sizes[filename] = st.st_size

        digests = self._get_digests()
        for filename, actual_size in sizes.items():
            size = digests.get(filename, {}).get("size")
            if size is None:
                continue
            if size != actual_size:
                return False

        # All files are present and sizes are ok. In this case the normal
        # fetch code will be skipped, so we need to generate equivalent
        # output here.
        if self.logfile is not None:
            f = io.open(
                _unicode_encode(
                    self.logfile, encoding=_encodings["fs"], errors="strict"
                ),
                mode="a",
                encoding=_encodings["content"],
                errors="backslashreplace",
            )
            for filename in uri_map:
                f.write(
                    _unicode_decode(
                        (" * %s size ;-) ..." % filename).ljust(73) + "[ ok ]\n"
                    )
                )
            f.close()

        return True

    def _pipe(self, fd_pipes):
        """When appropriate, use a pty so that fetcher progress bars,
        like wget has, will work properly."""
        if self.background or not sys.stdout.isatty():
            # When the output only goes to a log file,
            # there's no point in creating a pty.
            return os.pipe()
        stdout_pipe = None
        if not self.background:
            stdout_pipe = fd_pipes.get(1)
        got_pty, master_fd, slave_fd = _create_pty_or_pipe(copy_term_size=stdout_pipe)
        return (master_fd, slave_fd)

    def _eerror(self, lines):
        out = io.StringIO()
        for line in lines:
            eerror(line, phase="unpack", key=self.pkg.cpv, out=out)
        msg = out.getvalue()
        if msg:
            self.scheduler.output(msg, log_path=self.logfile)

    def _proc_join_done(self, proc, future):
        """
        Extend _proc_join_done to emit an eerror message for fetch failure.
        """
        if not self.prefetch and not future.cancelled() and proc.exitcode != os.EX_OK:
            msg_lines = []
            msg = "Fetch failed for '%s'" % (self.pkg.cpv,)
            if self.logfile is not None:
                msg += ", Log file:"
            msg_lines.append(msg)
            if self.logfile is not None:
                msg_lines.append(" '%s'" % (self.logfile,))
            self._eerror(msg_lines)
        super(_EbuildFetcherProcess, self)._proc_join_done(proc, future)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import io
import sys
import time

import portage
import portage.util.formatter as formatter
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.output import xtermTitle

from _emerge.getloadavg import getloadavg


class JobStatusDisplay:

    _bound_properties = ("curval", "failed", "running")

    # Don't update the display unless at least this much
    # time has passed, in units of seconds.
    _min_display_latency = 2

    _default_term_codes = {
        "cr": "\r",
        "el": "\x1b[K",
        "nel": "\n",
    }

    _termcap_name_map = {
        "carriage_return": "cr",
        "clr_eol": "el",
        "newline": "nel",
    }

    def __init__(self, quiet=False, xterm_titles=True):
        object.__setattr__(self, "quiet", quiet)
        object.__setattr__(self, "xterm_titles", xterm_titles)
        object.__setattr__(self, "maxval", 0)
        object.__setattr__(self, "merges", 0)
        object.__setattr__(self, "_changed", False)
        object.__setattr__(self, "_displayed", False)
        object.__setattr__(self, "_last_display_time", 0)

        self.reset()

        isatty = (
            os.environ.get("TERM") != "dumb"
            and hasattr(self.out, "isatty")
            and self.out.isatty()
        )
        object.__setattr__(self, "_isatty", isatty)
        if not isatty or not self._init_term():
            term_codes = {}
            for k, capname in self._termcap_name_map.items():
                term_codes[k] = self._default_term_codes[capname]
            object.__setattr__(self, "_term_codes", term_codes)
        encoding = sys.getdefaultencoding()
        for k, v in self._term_codes.items():
            if not isinstance(v, str):
                self._term_codes[k] = v.decode(encoding, "replace")

        if self._isatty:
            width = portage.output.get_term_size()[1]
        else:
            width = 80
        self._set_width(width)

    def _set_width(self, width):
        if width == getattr(self, "width", None):
            return
        if width <= 0 or width > 80:
            width = 80
        object.__setattr__(self, "width", width)
        object.__setattr__(self, "_jobs_column_width", width - 32)

    @property
    def out(self):
        """Use a lazy reference to sys.stdout, in case the API consumer has
        temporarily overridden stdout."""
        return sys.stdout

    def _write(self, s):
        # avoid potential UnicodeEncodeError
        s = _unicode_encode(s, encoding=_encodings["stdio"], errors="backslashreplace")
        out = self.out.buffer
        out.write(s)
        out.flush()

    def _init_term(self):
        """
        Initialize term control codes.
        @rtype: bool
        @return: True if term codes were successfully initialized,
                False otherwise.
        """

        term_type = os.environ.get("TERM", "").strip()
        if not term_type:
            return False
        tigetstr = None

        try:
            import curses

            try:
                curses.setupterm(term_type, self.out.fileno())
                tigetstr = curses.tigetstr
            except curses.error:
                pass
        except ImportError:
            pass

        if tigetstr is None:
            return False

        term_codes = {}
        for k, capname in self._termcap_name_map.items():
            # Use _native_string for PyPy compat (bug #470258).
            code = tigetstr(portage._native_string(capname))
            if code is None:
                code = self._default_term_codes[capname]
            term_codes[k] = code
        object.__setattr__(self, "_term_codes", term_codes)
        return True

    def _format_msg(self, msg):
        return ">>> %s" % msg

    def _erase(self):
        self._write(self._term_codes["carriage_return"] + self._term_codes["clr_eol"])
        self._displayed = False

    def _display(self, line):
        self._write(line)
        self._displayed = True

    def _update(self, msg):

        if not self._isatty:
            self._write(self._format_msg(msg) + self._term_codes["newline"])
            self._displayed = True
            return

        if self._displayed:
            self._erase()

        self._display(self._format_msg(msg))

    def displayMessage(self, msg):

        was_displayed = self._displayed

        if self._isatty and self._displayed:
            self._erase()

        self._write(self._format_msg(msg) + self._term_codes["newline"])
        self._displayed = False

        if was_displayed:
            self._changed = True
            self.display()

    def reset(self):
        self.maxval = 0
        self.merges = 0
        for name in self._bound_properties:
            object.__setattr__(self, name, 0)

        if self._displayed:
            self._write(self._term_codes["newline"])
            self._displayed = False

    def __setattr__(self, name, value):
        old_value = getattr(self, name)
        if value == old_value:
            return
        object.__setattr__(self, name, value)
        if name in self._bound_properties:
            self._property_change(name, old_value, value)

    def _property_change(self, name, old_value, new_value):
        self._changed = True
        self.display()

    def _load_avg_str(self):
        try:
            avg = getloadavg()
        except OSError:
            return "unknown"

        max_avg = max(avg)

        if max_avg < 10:
            digits = 2
        elif max_avg < 100:
            digits = 1
        else:
            digits = 0

        return ", ".join(("%%.%df" % digits) % x for x in avg)

    def display(self):
        """
        Display status on stdout, but only if something has
        changed since the last call. This always returns True,
        for continuous scheduling via timeout_add.
        """

        if self.quiet:
            return True

        current_time = time.time()
        time_delta = current_time - self._last_display_time
        if self._displayed and not self._changed:
            if not self._isatty:
                return True
            if time_delta < self._min_display_latency:
                return True

        self._last_display_time = current_time
        self._changed = False
        self._display_status()
        return True

    def _display_status(self):
        # Don't use len(self._completed_tasks) here since that also
        # can include uninstall tasks.
        curval_str = "%s" % (self.curval,)
        maxval_str = "%s" % (self.maxval,)
        running_str = "%s" % (self.running,)
        failed_str = "%s" % (self.failed,)
        load_avg_str = self._load_avg_str()

        color_output = io.StringIO()
        plain_output = io.StringIO()
        style_file = portage.output.ConsoleStyleFile(color_output)
        style_file.write_listener = plain_output
        style_writer = portage.output.StyleWriter(file=style_file, maxcol=9999)
        style_writer.style_listener = style_file.new_styles
        f = formatter.AbstractFormatter(style_writer)

        number_style = "INFORM"
        f.add_literal_data("Jobs: ")
        f.push_style(number_style)
        f.add_literal_data(curval_str)
        f.pop_style()
        f.add_literal_data(" of ")
        f.push_style(number_style)
        f.add_literal_data(maxval_str)
        f.pop_style()
        f.add_literal_data(" complete")

        if self.running:
            f.add_literal_data(", ")
            f.push_style(number_style)
            f.add_literal_data(running_str)
            f.pop_style()
            f.add_literal_data(" running")

        if self.failed:
            f.add_literal_data(", ")
            f.push_style(number_style)
            f.add_literal_data(failed_str)
            f.pop_style()
            f.add_literal_data(" failed")

        padding = self._jobs_column_width - len(plain_output.getvalue())
        if padding > 0:
            f.add_literal_data(padding * " ")

        f.add_literal_data("Load avg: ")
        f.add_literal_data(load_avg_str)

        # Truncate to fit width, to avoid making the terminal scroll if the
        # line overflows (happens when the load average is large).
        plain_output = plain_output.getvalue()
        if self._isatty and len(plain_output) > self.width:
            # Use plain_output here since it's easier to truncate
            # properly than the color output which contains console
            # color codes.
            self._update(plain_output[: self.width])
        else:
            self._update(color_output.getvalue())

        if self.xterm_titles:
            # If the HOSTNAME variable is exported, include it
            # in the xterm title, just like emergelog() does.
            # See bug #390699.
            title_str = " ".join(plain_output.split())
            hostname = os.environ.get("HOSTNAME")
            if hostname is not None:
                title_str = "%s: %s" % (hostname, title_str)
            xtermTitle(title_str)



# Copyright 2012-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import fcntl
import functools
import multiprocessing
import signal
import sys

import portage
from portage import os
from portage.util.futures import asyncio
from _emerge.SpawnProcess import SpawnProcess


class ForkProcess(SpawnProcess):

    __slots__ = ("_proc", "_proc_join_task")

    # Number of seconds between poll attempts for process exit status
    # (after the sentinel has become ready).
    _proc_join_interval = 0.1

    def _spawn(self, args, fd_pipes=None, **kwargs):
        """
        Override SpawnProcess._spawn to fork a subprocess that calls
        self._run(). This uses multiprocessing.Process in order to leverage
        any pre-fork and post-fork interpreter housekeeping that it provides,
        promoting a healthy state for the forked interpreter.
        """
        # Since multiprocessing.Process closes sys.__stdin__, create a
        # temporary duplicate of fd_pipes[0] so that sys.__stdin__ can
        # be restored in the subprocess, in case this is needed for
        # things like PROPERTIES=interactive support.
        stdin_dup = None
        try:
            stdin_fd = fd_pipes.get(0)
            if stdin_fd is not None and stdin_fd == portage._get_stdin().fileno():
                stdin_dup = os.dup(stdin_fd)
                fcntl.fcntl(
                    stdin_dup, fcntl.F_SETFD, fcntl.fcntl(stdin_fd, fcntl.F_GETFD)
                )
                fd_pipes[0] = stdin_dup
            self._proc = multiprocessing.Process(
                target=self._bootstrap, args=(fd_pipes,)
            )
            self._proc.start()
        finally:
            if stdin_dup is not None:
                os.close(stdin_dup)

        self._proc_join_task = asyncio.ensure_future(
            self._proc_join(self._proc, loop=self.scheduler), loop=self.scheduler
        )
        self._proc_join_task.add_done_callback(
            functools.partial(self._proc_join_done, self._proc)
        )

        return [self._proc.pid]

    def _cancel(self):
        if self._proc is None:
            super(ForkProcess, self)._cancel()
        else:
            self._proc.terminate()

    def _async_wait(self):
        if self._proc_join_task is None:
            super(ForkProcess, self)._async_wait()

    def _async_waitpid(self):
        if self._proc_join_task is None:
            super(ForkProcess, self)._async_waitpid()

    async def _proc_join(self, proc, loop=None):
        sentinel_reader = self.scheduler.create_future()
        self.scheduler.add_reader(
            proc.sentinel,
            lambda: sentinel_reader.done() or sentinel_reader.set_result(None),
        )
        try:
            await sentinel_reader
        finally:
            # If multiprocessing.Process supports the close method, then
            # access to proc.sentinel will raise ValueError if the
            # sentinel has been closed. In this case it's not safe to call
            # remove_reader, since the file descriptor may have been closed
            # and then reallocated to a concurrent coroutine. When the
            # close method is not supported, proc.sentinel remains open
            # until proc's finalizer is called.
            try:
                self.scheduler.remove_reader(proc.sentinel)
            except ValueError:
                pass

        # Now that proc.sentinel is ready, poll until process exit
        # status has become available.
        while True:
            proc.join(0)
            if proc.exitcode is not None:
                break
            await asyncio.sleep(self._proc_join_interval, loop=loop)

    def _proc_join_done(self, proc, future):
        future.cancelled() or future.result()
        self._was_cancelled()
        if self.returncode is None:
            self.returncode = proc.exitcode

        self._proc = None
        if hasattr(proc, "close"):
            proc.close()
        self._proc_join_task = None
        self._async_wait()

    def _unregister(self):
        super(ForkProcess, self)._unregister()
        if self._proc is not None:
            if self._proc.is_alive():
                self._proc.terminate()
            self._proc = None
        if self._proc_join_task is not None:
            self._proc_join_task.cancel()
            self._proc_join_task = None

    def _bootstrap(self, fd_pipes):
        # Use default signal handlers in order to avoid problems
        # killing subprocesses as reported in bug #353239.
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # Unregister SIGCHLD handler and wakeup_fd for the parent
        # process's event loop (bug 655656).
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        try:
            wakeup_fd = signal.set_wakeup_fd(-1)
            if wakeup_fd > 0:
                os.close(wakeup_fd)
        except (ValueError, OSError):
            pass

        portage.locks._close_fds()
        # We don't exec, so use close_fds=False
        # (see _setup_pipes docstring).
        portage.process._setup_pipes(fd_pipes, close_fds=False)

        # Since multiprocessing.Process closes sys.__stdin__ and
        # makes sys.stdin refer to os.devnull, restore it when
        # appropriate.
        if 0 in fd_pipes:
            # It's possible that sys.stdin.fileno() is already 0,
            # and in that case the above _setup_pipes call will
            # have already updated its identity via dup2. Otherwise,
            # perform the dup2 call now, and also copy the file
            # descriptor flags.
            if sys.stdin.fileno() != 0:
                os.dup2(0, sys.stdin.fileno())
                fcntl.fcntl(
                    sys.stdin.fileno(), fcntl.F_SETFD, fcntl.fcntl(0, fcntl.F_GETFD)
                )
            sys.__stdin__ = sys.stdin

        sys.exit(self._run())

    def _run(self):
        raise NotImplementedError(self)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging

from portage import os
from portage.util import writemsg_level
from portage.util.futures import asyncio
from _emerge.AbstractPollTask import AbstractPollTask
import signal
import errno


class SubProcess(AbstractPollTask):

    __slots__ = ("pid",) + ("_dummy_pipe_fd", "_files", "_waitpid_id")

    # This is how much time we allow for waitpid to succeed after
    # we've sent a kill signal to our subprocess.
    _cancel_timeout = 1  # seconds

    def _poll(self):
        # Simply rely on _async_waitpid_cb to set the returncode.
        return self.returncode

    def _cancel(self):
        if self.isAlive() and self.pid is not None:
            try:
                os.kill(self.pid, signal.SIGTERM)
            except OSError as e:
                if e.errno == errno.EPERM:
                    # Reported with hardened kernel (bug #358211).
                    writemsg_level(
                        "!!! kill: (%i) - Operation not permitted\n" % (self.pid,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                elif e.errno != errno.ESRCH:
                    raise

    def _async_wait(self):
        if self.returncode is None:
            raise asyncio.InvalidStateError("Result is not ready for %s" % (self,))
        else:
            # This calls _unregister, so don't call it until pid status
            # is available.
            super(SubProcess, self)._async_wait()

    def _async_waitpid(self):
        """
        Wait for exit status of self.pid asynchronously, and then
        set the returncode, and finally notify exit listeners via the
        _async_wait method. Subclasses may override this method in order
        to implement an alternative means to retrieve pid exit status,
        or as a means to delay action until some pending task(s) have
        completed (such as reading data that the subprocess is supposed
        to have written to a pipe).
        """
        if self.returncode is not None:
            self._async_wait()
        elif self._waitpid_id is None:
            self._waitpid_id = self.pid
            self.scheduler._asyncio_child_watcher.add_child_handler(
                self.pid, self._async_waitpid_cb
            )

    def _async_waitpid_cb(self, pid, returncode):
        if pid != self.pid:
            raise AssertionError("expected pid %s, got %s" % (self.pid, pid))
        self.returncode = returncode
        self._async_wait()

    def _orphan_process_warn(self):
        pass

    def _unregister(self):
        """
        Unregister from the scheduler and close open files.
        """

        self._registered = False

        if self._waitpid_id is not None:
            self.scheduler._asyncio_child_watcher.remove_child_handler(self._waitpid_id)
            self._waitpid_id = None

        if self._files is not None:
            for f in self._files.values():
                if isinstance(f, int):
                    os.close(f)
                else:
                    f.close()
            self._files = None



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.repository.storage.interface import (
    RepoStorageException,
    RepoStorageInterface,
)
from portage.util.futures import asyncio

from _emerge.SpawnProcess import SpawnProcess


class HardlinkQuarantineRepoStorage(RepoStorageInterface):
    """
    This is the default storage module, since its quite compatible with
    most configurations.

    It's desirable to be able to create shared hardlinks between the
    download directory and the normal repository, and this is facilitated
    by making the download directory be a subdirectory of the normal
    repository location (ensuring that no mountpoints are crossed).
    Shared hardlinks are created by using the rsync --link-dest option.

    Since the download is initially unverified, it is safest to save
    it in a quarantine directory. The quarantine directory is also
    useful for making the repository update more atomic, so that it
    less likely that normal repository location will be observed in
    a partially synced state.
    """

    def __init__(self, repo, spawn_kwargs):
        self._user_location = repo.location
        self._update_location = None
        self._spawn_kwargs = spawn_kwargs
        self._current_update = None

    async def _check_call(self, cmd):
        """
        Run cmd and raise RepoStorageException on failure.

        @param cmd: command to executre
        @type cmd: list
        """
        p = SpawnProcess(
            args=cmd, scheduler=asyncio.get_event_loop(), **self._spawn_kwargs
        )
        p.start()
        if await p.async_wait() != os.EX_OK:
            raise RepoStorageException(
                "command exited with status {}: {}".format(p.returncode, " ".join(cmd))
            )

    async def init_update(self):
        update_location = os.path.join(
            self._user_location, ".tmp-unverified-download-quarantine"
        )
        await self._check_call(["rm", "-rf", update_location])

        # Use  rsync --link-dest to hardlink a files into self._update_location,
        # since cp -l is not portable.
        await self._check_call(
            [
                "rsync",
                "-a",
                "--link-dest",
                self._user_location,
                "--exclude=/distfiles",
                "--exclude=/local",
                "--exclude=/lost+found",
                "--exclude=/packages",
                "--exclude",
                "/{}".format(os.path.basename(update_location)),
                self._user_location + "/",
                update_location + "/",
            ]
        )

        self._update_location = update_location

        return self._update_location

    @property
    def current_update(self):
        if self._update_location is None:
            raise RepoStorageException("current update does not exist")
        return self._update_location

    async def commit_update(self):
        update_location = self.current_update
        self._update_location = None
        await self._check_call(
            [
                "rsync",
                "-a",
                "--delete",
                "--exclude=/distfiles",
                "--exclude=/local",
                "--exclude=/lost+found",
                "--exclude=/packages",
                "--exclude",
                "/{}".format(os.path.basename(update_location)),
                update_location + "/",
                self._user_location + "/",
            ]
        )

        await self._check_call(["rm", "-rf", update_location])

    async def abort_update(self):
        if self._update_location is not None:
            update_location = self._update_location
            self._update_location = None
            await self._check_call(["rm", "-rf", update_location])

    async def garbage_collection(self):
        await self.abort_update()



# Copyright 2014-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections

from portage.versions import _pkg_str


pkg_desc_index_node = collections.namedtuple(
    "pkg_desc_index_node", ["cp", "cpv_list", "desc"]
)


class pkg_node(str):
    """
    A minimal package node class. For performance reasons, inputs
    are not validated.
    """

    def __init__(self, cp, version, repo=None):
        self.__dict__["cp"] = cp
        self.__dict__["repo"] = repo
        self.__dict__["version"] = version
        self.__dict__["build_time"] = None

    def __new__(cls, cp, version, repo=None):
        return str.__new__(cls, cp + "-" + version)

    def __setattr__(self, name, value):
        raise AttributeError(
            "pkg_node instances are immutable", self.__class__, name, value
        )


def pkg_desc_index_line_format(cp, pkgs, desc):
    return "%s %s: %s\n" % (cp, " ".join(_pkg_str(cpv).version for cpv in pkgs), desc)


def pkg_desc_index_line_read(line, repo=None):

    try:
        pkgs, desc = line.split(":", 1)
    except ValueError:
        return None
    desc = desc.strip()

    try:
        cp, pkgs = pkgs.split(" ", 1)
    except ValueError:
        return None

    cp_list = []
    for ver in pkgs.split():
        cp_list.append(pkg_node(cp, ver, repo))

    return pkg_desc_index_node(cp, tuple(cp_list), desc)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


import portage
from portage import os
from portage import digraph
from portage._sets.base import InternalPackageSet
from portage.dep import Atom

from _emerge.BlockerCache import BlockerCache
from _emerge.Package import Package
from _emerge.show_invalid_depstring_notice import show_invalid_depstring_notice


class BlockerDB:
    def __init__(self, fake_vartree):
        root_config = fake_vartree._root_config
        self._root_config = root_config
        self._vartree = root_config.trees["vartree"]
        self._portdb = root_config.trees["porttree"].dbapi

        self._dep_check_trees = None
        self._fake_vartree = fake_vartree
        self._dep_check_trees = {
            self._vartree.settings["EROOT"]: {
                "porttree": fake_vartree,
                "vartree": fake_vartree,
            }
        }

    def findInstalledBlockers(self, new_pkg):
        """
        Search for installed run-time blockers in the root where
        new_pkg is planned to be installed. This ignores build-time
        blockers, since new_pkg is assumed to be built already.
        """
        blocker_cache = BlockerCache(None, self._vartree.dbapi)
        dep_keys = Package._runtime_keys
        settings = self._vartree.settings
        stale_cache = set(blocker_cache)
        fake_vartree = self._fake_vartree
        dep_check_trees = self._dep_check_trees
        vardb = fake_vartree.dbapi
        installed_pkgs = list(vardb)

        for inst_pkg in installed_pkgs:
            stale_cache.discard(inst_pkg.cpv)
            cached_blockers = blocker_cache.get(inst_pkg.cpv)
            if (
                cached_blockers is not None
                and cached_blockers.counter != inst_pkg.counter
            ):
                cached_blockers = None
            if cached_blockers is not None:
                blocker_atoms = cached_blockers.atoms
            else:
                # Use aux_get() to trigger FakeVartree global
                # updates on *DEPEND when appropriate.
                depstr = " ".join(vardb.aux_get(inst_pkg.cpv, dep_keys))
                success, atoms = portage.dep_check(
                    depstr,
                    vardb,
                    settings,
                    myuse=inst_pkg.use.enabled,
                    trees=dep_check_trees,
                    myroot=inst_pkg.root,
                )
                if not success:
                    pkg_location = os.path.join(
                        inst_pkg.root, portage.VDB_PATH, inst_pkg.category, inst_pkg.pf
                    )
                    portage.writemsg(
                        "!!! %s/*DEPEND: %s\n" % (pkg_location, atoms), noiselevel=-1
                    )
                    continue

                blocker_atoms = [atom for atom in atoms if atom.startswith("!")]
                blocker_atoms.sort()
                blocker_cache[inst_pkg.cpv] = blocker_cache.BlockerData(
                    inst_pkg.counter, blocker_atoms
                )
        for cpv in stale_cache:
            del blocker_cache[cpv]
        blocker_cache.flush()

        blocker_parents = digraph()
        blocker_atoms = []
        for pkg in installed_pkgs:
            for blocker_atom in blocker_cache[pkg.cpv].atoms:
                blocker_atom = blocker_atom.lstrip("!")
                blocker_atoms.append(blocker_atom)
                blocker_parents.add(blocker_atom, pkg)

        blocker_atoms = InternalPackageSet(initial_atoms=blocker_atoms)
        blocking_pkgs = set()
        for atom in blocker_atoms.iterAtomsForPackage(new_pkg):
            blocking_pkgs.update(blocker_parents.parent_nodes(atom))

        # Check for blockers in the other direction.
        depstr = " ".join(new_pkg._metadata[k] for k in dep_keys)
        success, atoms = portage.dep_check(
            depstr,
            vardb,
            settings,
            myuse=new_pkg.use.enabled,
            trees=dep_check_trees,
            myroot=new_pkg.root,
        )
        if not success:
            # We should never get this far with invalid deps.
            show_invalid_depstring_notice(new_pkg, atoms)
            assert False

        blocker_atoms = [atom.lstrip("!") for atom in atoms if atom[:1] == "!"]
        if blocker_atoms:
            blocker_atoms = InternalPackageSet(initial_atoms=blocker_atoms)
            for inst_pkg in installed_pkgs:
                try:
                    next(blocker_atoms.iterAtomsForPackage(inst_pkg))
                except (portage.exception.InvalidDependString, StopIteration):
                    continue
                blocking_pkgs.add(inst_pkg)

        return blocking_pkgs

    def discardBlocker(self, pkg):
        """Discard a package from the list of potential blockers.
        This will match any package(s) with identical cpv or cp:slot."""
        for cpv_match in self._fake_vartree.dbapi.match_pkgs(Atom("=%s" % (pkg.cpv,))):
            if cpv_match.cp == pkg.cp:
                self._fake_vartree.cpv_discard(cpv_match)
        for slot_match in self._fake_vartree.dbapi.match_pkgs(pkg.slot_atom):
            if slot_match.cp == pkg.cp:
                self._fake_vartree.cpv_discard(slot_match)



#!/usr/bin/env python
# Copyright 2017-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
#
# Copyright 2017 The Chromium OS Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Core implementation of doins ebuild helper command.

This script is designed to be executed by ebuild-helpers/doins.
"""

import argparse
import errno
import grp
import logging
import os
import pwd
import shlex
import shutil
import stat
import subprocess
import sys

from portage.util import movefile
from portage.util.file_copy import copyfile

logger = logging.getLogger("portage.bin.doins")


def _warn(helper, msg):
    """Output warning message to stderr.

    Args:
            helper: helper executable name.
            msg: Message to be output.
    """
    print("!!! %s: %s\n" % (helper, msg), file=sys.stderr)


def _parse_group(group):
    """Parses gid.

    Args:
            group: string representation of the group. Maybe name or gid.
    Returns:
            Parsed gid.
    """
    try:
        return grp.getgrnam(group).gr_gid
    except KeyError:
        pass
    return int(group)


def _parse_user(user):
    """Parses uid.

    Args:
            user: string representation of the user. Maybe name or uid.
    Returns:
            Parsed uid.
    """
    try:
        return pwd.getpwnam(user).pw_uid
    except KeyError:
        pass
    return int(user)


def _parse_mode(mode):
    """Parses mode.

    Args:
            mode: string representation of the permission.
    Returns:
            Parsed mode.
    """
    # `install`'s --mode option is complicated. So here is partially
    # supported.
    try:
        return int(mode, 8)
    except ValueError:
        # In case of fail, returns None, so that caller can check
        # if unknown '-m' is set or not.
        return None


def _parse_install_options(
    options, is_strict, helper, inprocess_runner_class, subprocess_runner_class
):
    """Parses command line arguments for `install` command.

    Args:
            options: string representation of `install` options.
            is_strict: bool. If True, this exits the program in case of
                    that an unknown option is found.
            helper: helper executable name.
            inprocess_runner_class: Constructor to run procedure which
                    `install` command will do.
            subprocess_runner_class: Constructor to run `install` command.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-g", "--group", default=-1, type=_parse_group)
    parser.add_argument("-o", "--owner", default=-1, type=_parse_user)
    parser.add_argument("-m", "--mode", default=0o755, type=_parse_mode)
    parser.add_argument("-p", "--preserve-timestamps", action="store_true")
    split_options = shlex.split(options)
    namespace, remaining = parser.parse_known_args(split_options)
    # Because parsing '--mode' option is partially supported. If unknown
    # arg for --mode is passed, namespace.mode is set to None.
    if remaining or namespace.mode is None:
        _warn(helper, "Unknown install options: %s, %r" % (options, remaining))
        if is_strict:
            sys.exit(1)
        _warn(
            helper,
            "Continue with falling back to `install` "
            "command execution, which can be slower.",
        )
        return subprocess_runner_class(split_options)
    return inprocess_runner_class(namespace)


def _set_attributes(options, path):
    """Sets attributes the file/dir at given |path|.

    Args:
            options: object which has |owner|, |group| and |mode| fields.
                    |owner| is int value representing uid. Similary |group|
                    represents gid.
                    If -1 is set, just unchanged.
                    |mode| is the bits of permissions.
            path: File/directory path.
    """
    if options.owner != -1 or options.group != -1:
        os.lchown(path, options.owner, options.group)
    if options.mode is not None:
        os.chmod(path, options.mode)


def _set_timestamps(source_stat, dest):
    """Apply timestamps from source_stat to dest.

    Args:
            source_stat: stat result for the source file.
            dest: path to the dest file.
    """
    os.utime(dest, ns=(source_stat.st_atime_ns, source_stat.st_mtime_ns))


class _InsInProcessInstallRunner:
    """Implements `install` command behavior running in a process."""

    def __init__(self, opts, parsed_options):
        """Initializes the instance.

        Args:
                opts: namespace object containing the parsed
                        arguments for this program.
                parsed_options: namespace object contaning the parsed
                        options for `install`.
        """
        self._parsed_options = parsed_options
        self._helper = opts.helper
        self._copy_xattr = opts.enable_copy_xattr
        if self._copy_xattr:
            self._xattr_exclude = opts.xattr_exclude

    def run(self, source, dest_dir):
        """Installs a file at |source| into |dest_dir| in process.

        Args:
                source: Path to the file to be installed.
                dest_dir: Path to the directory which |source| will be
                        installed into.
        Returns:
                True on success, otherwise False.
        """
        dest = os.path.join(dest_dir, os.path.basename(source))
        # Raise an exception if stat(source) fails, intentionally.
        sstat = os.stat(source)
        if not self._is_install_allowed(source, sstat, dest):
            return False

        # To emulate the `install` command, remove the dest file in
        # advance.
        try:
            os.unlink(dest)
        except OSError as e:
            # Removing a non-existing entry should be handled as a
            # regular case.
            if e.errno != errno.ENOENT:
                raise
        try:
            copyfile(source, dest)
            _set_attributes(self._parsed_options, dest)
            if self._copy_xattr:
                movefile._copyxattr(source, dest, exclude=self._xattr_exclude)
            if self._parsed_options.preserve_timestamps:
                _set_timestamps(sstat, dest)
        except Exception:
            logger.exception(
                "Failed to copy file: " "_parsed_options=%r, source=%r, dest_dir=%r",
                self._parsed_options,
                source,
                dest_dir,
            )
            return False
        return True

    def _is_install_allowed(self, source, source_stat, dest):
        """Returns if installing source into dest should work.

        This is to keep compatibility with the `install` command.

        Args:
                source: path to the source file.
                source_stat: stat result for the source file, using stat()
                        rather than lstat(), in order to match the `install`
                        command
                dest: path to the dest file.

        Returns:
                True if it should succeed.
        """
        # To match `install` command, use stat() for source, while
        # lstat() for dest.
        try:
            dest_lstat = os.lstat(dest)
        except OSError as e:
            # It is common to install a file into a new path,
            # so if the destination doesn't exist, ignore it.
            if e.errno == errno.ENOENT:
                return True
            raise

        # Allowing install, if the target is a symlink.
        if stat.S_ISLNK(dest_lstat.st_mode):
            return True

        # Allowing install, if source file and dest file are different.
        # Note that, later, dest will be unlinked.
        if not os.path.samestat(source_stat, dest_lstat):
            return True

        # Allowing install, in hardlink case, if the actual path are
        # different, because source can be preserved even after dest is
        # unlinked.
        if dest_lstat.st_nlink > 1 and os.path.realpath(source) != os.path.realpath(
            dest
        ):
            return True

        _warn(self._helper, "%s and %s are same file." % (source, dest))
        return False


class _InsSubprocessInstallRunner:
    """Runs `install` command in a subprocess to install a file."""

    def __init__(self, split_options):
        """Initializes the instance.

        Args:
                split_options: Command line options to be passed to
                        `install` command. List of str.
        """
        self._split_options = split_options

    def run(self, source, dest_dir):
        """Installs a file at |source| into |dest_dir| by `install`.

        Args:
                source: Path to the file to be installed.
                dest_dir: Path to the directory which |source| will be
                installed into.
        Returns:
                True on success, otherwise False.
        """
        command = ["install"] + self._split_options + [source, dest_dir]
        return subprocess.call(command) == 0


class _DirInProcessInstallRunner:
    """Implements `install` command behavior running in a process."""

    def __init__(self, parsed_options):
        """Initializes the instance.

        Args:
                parsed_options: namespace object contaning the parsed
                        options for `install`.
        """
        self._parsed_options = parsed_options

    def run(self, dest):
        """Installs a dir into |dest| in process.

        Args:
                dest: Path where a directory should be created.
        """
        try:
            os.makedirs(dest)
        except OSError as e:
            if e.errno != errno.EEXIST or not os.path.isdir(dest):
                raise
        _set_attributes(self._parsed_options, dest)


class _DirSubprocessInstallRunner:
    """Runs `install` command to create a directory."""

    def __init__(self, split_options):
        """Initializes the instance.

        Args:
                split_options: Command line options to be passed to
                        `install` command. List of str.
        """
        self._split_options = split_options

    def run(self, dest):
        """Installs a dir into |dest| by `install` command.

        Args:
                dest: Path where a directory should be created.
        """
        command = ["install", "-d"] + self._split_options + [dest]
        subprocess.check_call(command)


class _InstallRunner:
    """Handles `install` command operation.

    Runs operations which `install` command should work. If possible,
    this may just call in-process functions, instead of executing `install`
    in a subprocess for performance.
    """

    def __init__(self, opts):
        """Initializes the instance.

        Args:
                opts: namespace object containing the parsed
                        arguments for this program.
        """
        self._ins_runner = _parse_install_options(
            opts.insoptions,
            opts.strict_option,
            opts.helper,
            lambda options: _InsInProcessInstallRunner(opts, options),
            _InsSubprocessInstallRunner,
        )
        self._dir_runner = _parse_install_options(
            opts.diroptions,
            opts.strict_option,
            opts.helper,
            _DirInProcessInstallRunner,
            _DirSubprocessInstallRunner,
        )
        self._helpers_can_die = opts.helpers_can_die

    def install_file(self, source, dest_dir):
        """Installs a file at |source| into |dest_dir| directory.

        Args:
                source: Path to the file to be installed.
                dest_dir: Path to the directory which |source| will be
                        installed into.
        Returns:
                True on success, otherwise False.
        """
        return self._ins_runner.run(source, dest_dir)

    def install_dir(self, dest):
        """Creates a directory at |dest|.

        Args:
                dest: Path where a directory should be created.
        """
        try:
            self._dir_runner.run(dest)
        except Exception:
            if self._helpers_can_die:
                raise
            logger.exception("install_dir failed.")


def _doins(opts, install_runner, relpath, source_root):
    """Installs a file as if `install` command runs.

    Installs a file at |source_root|/|relpath| into
    |opts.dest|/|relpath|.
    If |args.preserve_symlinks| is set, creates symlink if the source is a
    symlink.

    Args:
            opts: parsed arguments. It should have following fields.
                    - preserve_symlinks: bool representing whether symlinks
                            needs to be preserved.
                    - dest: Destination root directory.
                    - distdir: location where Portage stores the downloaded
                            source code archives.
            install_runner: _InstallRunner instance for file install.
            relpath: Relative path of the file being installed.
            source_root: Source root directory.

    Returns: True on success.
    """
    source = os.path.join(source_root, relpath)
    dest = os.path.join(opts.dest, relpath)
    if os.path.islink(source):
        # Our fake ${DISTDIR} contains symlinks that should not be
        # reproduced inside ${D}. In order to ensure that things like
        # dodoc "${DISTDIR}"/foo.pdf work as expected, we dereference
        # symlinked files that refer to absolute paths inside
        # ${PORTAGE_ACTUAL_DISTDIR}/.
        try:
            if opts.preserve_symlinks and not os.readlink(source).startswith(
                opts.distdir
            ):
                linkto = os.readlink(source)
                try:
                    os.unlink(dest)
                except OSError as e:
                    if e.errno == errno.EISDIR:
                        shutil.rmtree(dest, ignore_errors=True)
                os.symlink(linkto, dest)
                return True
        except Exception:
            logger.exception(
                "Failed to create symlink: " "opts=%r, relpath=%r, source_root=%r",
                opts,
                relpath,
                source_root,
            )
            return False

    return install_runner.install_file(source, os.path.dirname(dest))


def _create_arg_parser():
    """Returns the parser for the command line arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--recursive",
        action="store_true",
        help="If set, installs files recursively. Otherwise, "
        "just skips directories.",
    )
    parser.add_argument(
        "--preserve_symlinks",
        action="store_true",
        help="If set, a symlink will be installed as symlink.",
    )
    parser.add_argument(
        "--helpers_can_die",
        action="store_true",
        help="If set, die in isolated-functions.sh is enabled. "
        "Specifically this is used to keep compatible dodir's "
        "behavior.",
    )
    parser.add_argument("--distdir", default="", help="Path to the actual distdir.")
    parser.add_argument(
        "--insoptions",
        default="",
        help="Options passed to `install` command for installing a " "file.",
    )
    parser.add_argument(
        "--diroptions",
        default="",
        help="Options passed to `install` command for installing a " "dir.",
    )
    parser.add_argument(
        "--strict_option",
        action="store_true",
        help="If set True, abort if insoptions/diroptions contains an "
        "option which cannot be interpreted by this script, instead of "
        "fallback to execute `install` command.",
    )
    parser.add_argument(
        "--enable_copy_xattr", action="store_true", help="Copies xattrs, if set True"
    )
    parser.add_argument(
        "--xattr_exclude",
        default="",
        help="White space delimited glob pattern to exclude xattr copy."
        "Used only if --enable_xattr_copy is set.",
    )

    # If helper is dodoc, it changes the behavior for the directory
    # install without --recursive.
    parser.add_argument("--helper", help="Name of helper.")
    parser.add_argument("--dest", help="Destination where the files are installed.")
    parser.add_argument(
        "sources", nargs="*", help="Source file/directory paths to be installed."
    )

    return parser


def _parse_args(argv):
    """Parses the command line arguments.

    Args:
            argv: command line arguments to be parsed.
    Returns:
            namespace instance containing the parsed argument data.
    """
    parser = _create_arg_parser()
    opts = parser.parse_args(argv)

    # Encode back to the original byte stream. Please see
    # http://bugs.python.org/issue8776.
    opts.distdir = os.fsencode(opts.distdir) + b"/"
    opts.dest = os.fsencode(opts.dest)
    opts.sources = [os.fsencode(source) for source in opts.sources]

    return opts


def _install_dir(opts, install_runner, source):
    """Installs directory at |source|.

    Args:
            opts: namespace instance containing parsed command line
                    argument data.
            install_runner: _InstallRunner instance for dir install.
            source: Path to the source directory.
    Returns:
            True on success, False on failure, or None on skipped.
    """
    if not opts.recursive:
        if opts.helper == "dodoc":
            _warn(opts.helper, "%s is a directory" % (source,))
            return False
        # Neither success nor fail. Return None to indicate skipped.
        return None

    # Strip trailing '/'s.
    source = source.rstrip(b"/")
    source_root = os.path.dirname(source)
    dest_dir = os.path.join(opts.dest, os.path.basename(source))
    install_runner.install_dir(dest_dir)

    relpath_list = []
    for dirpath, dirnames, filenames in os.walk(source):
        for dirname in dirnames:
            source_dir = os.path.join(dirpath, dirname)
            relpath = os.path.relpath(source_dir, source_root)
            if os.path.islink(source_dir):
                # If this is a symlink, it will be processed
                # in _doins() called later.
                relpath_list.append(relpath)
            else:
                dest = os.path.join(opts.dest, relpath)
                install_runner.install_dir(dest)
        relpath_list.extend(
            os.path.relpath(os.path.join(dirpath, filename), source_root)
            for filename in filenames
        )

    if not relpath_list:
        # NOTE: Even if only an empty directory is installed here, it
        # still counts as success, since an empty directory given as
        # an argument to doins -r should not trigger failure.
        return True
    success = True
    for relpath in relpath_list:
        if not _doins(opts, install_runner, relpath, source_root):
            success = False
    return success


def main(argv):
    opts = _parse_args(argv)
    install_runner = _InstallRunner(opts)

    if not os.path.isdir(opts.dest):
        install_runner.install_dir(opts.dest)

    any_success = False
    any_failure = False
    for source in opts.sources:
        if os.path.isdir(source) and (
            not opts.preserve_symlinks or not os.path.islink(source)
        ):
            ret = _install_dir(opts, install_runner, source)
            if ret is None:
                continue
            if ret:
                any_success = True
            else:
                any_failure = True
        else:
            if _doins(
                opts, install_runner, os.path.basename(source), os.path.dirname(source)
            ):
                any_success = True
            else:
                any_failure = True

    return 0 if not any_failure and any_success else 1


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python

# Ripped from HP and updated from Debian
# Update by Gentoo to support unicode output

#
# Copyright (c) 2004 Hewlett-Packard Development Company, L.P.
# 	David Mosberger <davidm@hpl.hp.com>
#
# Scan standard input for GCC warning messages that are likely to
# source of real 64-bit problems.  In particular, see whether there
# are any implicitly declared functions whose return values are later
# interpreted as pointers.  Those are almost guaranteed to cause
# crashes.
#

import re
import sys

implicit_pattern = re.compile(
    r"([^:]*):(\d+): warning: implicit declaration " + "of function [`']([^']*)'"
)
pointer_pattern = (
    r"([^:]*):(\d+): warning: "
    + r"("
    + r"(assignment"
    + r"|initialization"
    + r"|return"
    + r"|passing arg \d+ of `[^']*'"
    + r"|passing arg \d+ of pointer to function"
    + r") makes pointer from integer without a cast"
    + r"|"
    + r"cast to pointer from integer of different size)"
)

unicode_quote_open = "\u2018"
unicode_quote_close = "\u2019"


def write(msg):
    sys.stdout.buffer.write(msg.encode("utf_8", "backslashreplace"))


pointer_pattern = re.compile(pointer_pattern)

last_implicit_filename = ""
last_implicit_linenum = -1
last_implicit_func = ""

while True:
    line = sys.stdin.buffer.readline().decode("utf_8", "replace")
    if not line:
        break
    # translate unicode open/close quotes to ascii ones
    line = line.replace(unicode_quote_open, "`")
    line = line.replace(unicode_quote_close, "'")
    m = implicit_pattern.match(line)
    if m:
        last_implicit_filename = m.group(1)
        last_implicit_linenum = int(m.group(2))
        last_implicit_func = m.group(3)
    else:
        m = pointer_pattern.match(line)
        if m:
            pointer_filename = m.group(1)
            pointer_linenum = int(m.group(2))
            if (
                last_implicit_filename == pointer_filename
                and last_implicit_linenum == pointer_linenum
            ):
                write(
                    "Function `%s' implicitly converted to pointer at "
                    "%s:%d\n"
                    % (
                        last_implicit_func,
                        last_implicit_filename,
                        last_implicit_linenum,
                    )
                )



# Copyright 1999-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import gzip
import io
import sys
import tempfile

from _emerge.AsynchronousLock import AsynchronousLock
from _emerge.BinpkgEnvExtractor import BinpkgEnvExtractor
from _emerge.MiscFunctionsProcess import MiscFunctionsProcess
from _emerge.EbuildProcess import EbuildProcess
from _emerge.CompositeTask import CompositeTask
from _emerge.PackagePhase import PackagePhase
from _emerge.TaskSequence import TaskSequence
from portage.package.ebuild._ipc.QueryCommand import QueryCommand
from portage.util._dyn_libs.soname_deps_qa import (
    _get_all_provides,
    _get_unresolved_soname_deps,
)
from portage.package.ebuild.prepare_build_dirs import (
    _prepare_workdir,
    _prepare_fake_distdir,
    _prepare_fake_filesdir,
)
from portage.util import writemsg, ensure_dirs
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.BuildLogger import BuildLogger
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor
from portage.exception import InvalidBinaryPackageFormat
from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS

try:
    from portage.xml.metadata import MetaDataXML
except (SystemExit, KeyboardInterrupt):
    raise
except (ImportError, SystemError, RuntimeError, Exception):
    # broken or missing xml support
    # https://bugs.python.org/issue14988
    MetaDataXML = None

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.elog:messages@elog_messages",
    "portage.package.ebuild.doebuild:_check_build_log,"
    + "_post_phase_cmds,_post_phase_userpriv_perms,"
    + "_post_phase_emptydir_cleanup,"
    + "_post_src_install_soname_symlinks,"
    + "_post_src_install_uid_fix,_postinst_bsdflags,"
    + "_post_src_install_write_metadata,"
    + "_preinst_bsdflags",
    "portage.util.futures.unix_events:_set_nonblocking",
)
from portage import os
from portage import _encodings
from portage import _unicode_encode


class EbuildPhase(CompositeTask):

    __slots__ = ("actionmap", "fd_pipes", "phase", "settings") + ("_ebuild_lock",)

    # FEATURES displayed prior to setup phase
    _features_display = (
        "ccache",
        "compressdebug",
        "distcc",
        "fakeroot",
        "installsources",
        "keeptemp",
        "keepwork",
        "network-sandbox",
        "network-sandbox-proxy",
        "nostrip",
        "preserve-libs",
        "sandbox",
        "selinux",
        "sesandbox",
        "splitdebug",
        "suidctl",
        "test",
        "userpriv",
        "usersandbox",
    )

    # Locked phases
    _locked_phases = ("setup", "preinst", "postinst", "prerm", "postrm")

    def _start(self):
        future = asyncio.ensure_future(self._async_start(), loop=self.scheduler)
        self._start_task(AsyncTaskFuture(future=future), self._async_start_exit)

    async def _async_start(self):

        need_builddir = self.phase not in EbuildProcess._phases_without_builddir

        if need_builddir:
            phase_completed_file = os.path.join(
                self.settings["PORTAGE_BUILDDIR"], ".%sed" % self.phase.rstrip("e")
            )
            if not os.path.exists(phase_completed_file):
                # If the phase is really going to run then we want
                # to eliminate any stale elog messages that may
                # exist from a previous run.
                try:
                    os.unlink(os.path.join(self.settings["T"], "logging", self.phase))
                except OSError:
                    pass
            ensure_dirs(os.path.join(self.settings["PORTAGE_BUILDDIR"], "empty"))

        if self.phase in ("nofetch", "pretend", "setup"):

            use = self.settings.get("PORTAGE_BUILT_USE")
            if use is None:
                use = self.settings["PORTAGE_USE"]

            maint_str = ""
            upstr_str = ""
            metadata_xml_path = os.path.join(
                os.path.dirname(self.settings["EBUILD"]), "metadata.xml"
            )
            if MetaDataXML is not None and os.path.isfile(metadata_xml_path):
                herds_path = os.path.join(
                    self.settings["PORTDIR"], "metadata/herds.xml"
                )
                try:
                    metadata_xml = MetaDataXML(metadata_xml_path, herds_path)
                    maint_str = metadata_xml.format_maintainer_string()
                    upstr_str = metadata_xml.format_upstream_string()
                except SyntaxError:
                    maint_str = "<invalid metadata.xml>"

            msg = []
            msg.append(
                "Package:    %s:%s" % (self.settings.mycpv, self.settings["SLOT"])
            )
            if self.settings.get("PORTAGE_REPO_NAME"):
                msg.append("Repository: %s" % self.settings["PORTAGE_REPO_NAME"])
            if maint_str:
                msg.append("Maintainer: %s" % maint_str)
            if upstr_str:
                msg.append("Upstream:   %s" % upstr_str)

            msg.append("USE:        %s" % use)
            relevant_features = []
            enabled_features = self.settings.features
            for x in self._features_display:
                if x in enabled_features:
                    relevant_features.append(x)
            if relevant_features:
                msg.append("FEATURES:   %s" % " ".join(relevant_features))

            # Force background=True for this header since it's intended
            # for the log and it doesn't necessarily need to be visible
            # elsewhere.
            await self._elog("einfo", msg, background=True)

        if self.phase == "package":
            if "PORTAGE_BINPKG_TMPFILE" not in self.settings:
                binpkg_format = self.settings.get(
                    "BINPKG_FORMAT", SUPPORTED_GENTOO_BINPKG_FORMATS[0]
                )
                if binpkg_format == "xpak":
                    self.settings["BINPKG_FORMAT"] = "xpak"
                    self.settings["PORTAGE_BINPKG_TMPFILE"] = (
                        os.path.join(
                            self.settings["PKGDIR"],
                            self.settings["CATEGORY"],
                            self.settings["PF"],
                        )
                        + ".tbz2"
                    )
                elif binpkg_format == "gpkg":
                    self.settings["BINPKG_FORMAT"] = "gpkg"
                    self.settings["PORTAGE_BINPKG_TMPFILE"] = (
                        os.path.join(
                            self.settings["PKGDIR"],
                            self.settings["CATEGORY"],
                            self.settings["PF"],
                        )
                        + ".gpkg.tar"
                    )
                else:
                    raise InvalidBinaryPackageFormat(binpkg_format)

    def _async_start_exit(self, task):
        task.future.cancelled() or task.future.result()
        if self._default_exit(task) != os.EX_OK:
            self.wait()
            return

        if self.phase in ("pretend", "prerm"):
            env_extractor = BinpkgEnvExtractor(
                background=self.background,
                scheduler=self.scheduler,
                settings=self.settings,
            )
            if env_extractor.saved_env_exists():
                self._start_task(env_extractor, self._env_extractor_exit)
                return
            # If the environment.bz2 doesn't exist, then ebuild.sh will
            # source the ebuild as a fallback.

        self._start_lock()

    def _env_extractor_exit(self, env_extractor):
        if self._default_exit(env_extractor) != os.EX_OK:
            self.wait()
            return

        self._start_lock()

    def _start_lock(self):
        if (
            self.phase in self._locked_phases
            and "ebuild-locks" in self.settings.features
        ):
            eroot = self.settings["EROOT"]
            lock_path = os.path.join(eroot, portage.VDB_PATH + "-ebuild")
            if os.access(os.path.dirname(lock_path), os.W_OK):
                self._ebuild_lock = AsynchronousLock(
                    path=lock_path, scheduler=self.scheduler
                )
                self._start_task(self._ebuild_lock, self._lock_exit)
                return

        self._start_ebuild()

    def _lock_exit(self, ebuild_lock):
        if self._default_exit(ebuild_lock) != os.EX_OK:
            self.wait()
            return
        self._start_ebuild()

    def _get_log_path(self):
        # Don't open the log file during the clean phase since the
        # open file can result in an nfs lock on $T/build.log which
        # prevents the clean phase from removing $T.
        logfile = None
        if (
            self.phase not in ("clean", "cleanrm")
            and self.settings.get("PORTAGE_BACKGROUND") != "subprocess"
        ):
            logfile = self.settings.get("PORTAGE_LOG_FILE")
        return logfile

    def _start_ebuild(self):
        if self.phase == "package":
            self._start_task(
                PackagePhase(
                    actionmap=self.actionmap,
                    background=self.background,
                    fd_pipes=self.fd_pipes,
                    logfile=self._get_log_path(),
                    scheduler=self.scheduler,
                    settings=self.settings,
                ),
                self._ebuild_exit,
            )
            return

        if self.phase == "unpack":
            alist = self.settings.configdict["pkg"].get("A", "").split()
            _prepare_fake_distdir(self.settings, alist)
            _prepare_fake_filesdir(self.settings)

        fd_pipes = self.fd_pipes
        if fd_pipes is None:
            if not self.background and self.phase == "nofetch":
                # All the pkg_nofetch output goes to stderr since
                # it's considered to be an error message.
                fd_pipes = {1: sys.__stderr__.fileno()}

        ebuild_process = EbuildProcess(
            actionmap=self.actionmap,
            background=self.background,
            fd_pipes=fd_pipes,
            logfile=self._get_log_path(),
            phase=self.phase,
            scheduler=self.scheduler,
            settings=self.settings,
        )

        self._start_task(ebuild_process, self._ebuild_exit)

    def _ebuild_exit(self, ebuild_process):
        self._assert_current(ebuild_process)
        if self._ebuild_lock is None:
            self._ebuild_exit_unlocked(ebuild_process)
        else:
            self._start_task(
                AsyncTaskFuture(future=self._ebuild_lock.async_unlock()),
                functools.partial(self._ebuild_exit_unlocked, ebuild_process),
            )

    def _ebuild_exit_unlocked(self, ebuild_process, unlock_task=None):
        if unlock_task is not None:
            self._assert_current(unlock_task)
            if unlock_task.cancelled:
                self._default_final_exit(unlock_task)
                return

            # Normally, async_unlock should not raise an exception here.
            unlock_task.future.result()

        fail = False
        if ebuild_process.returncode != os.EX_OK:
            self.returncode = ebuild_process.returncode
            if self.phase == "test" and "test-fail-continue" in self.settings.features:
                # mark test phase as complete (bug #452030)
                try:
                    open(
                        _unicode_encode(
                            os.path.join(self.settings["PORTAGE_BUILDDIR"], ".tested"),
                            encoding=_encodings["fs"],
                            errors="strict",
                        ),
                        "wb",
                    ).close()
                except OSError:
                    pass
            else:
                fail = True

        if not fail:
            self.returncode = None

        logfile = self._get_log_path()

        if self.phase == "install":
            out = io.StringIO()
            _check_build_log(self.settings, out=out)
            msg = out.getvalue()
            self.scheduler.output(msg, log_path=logfile)

        if fail:
            self._die_hooks()
            return

        settings = self.settings
        _post_phase_userpriv_perms(settings)
        _post_phase_emptydir_cleanup(settings)

        if self.phase == "unpack":
            # Bump WORKDIR timestamp, in case tar gave it a timestamp
            # that will interfere with distfiles / WORKDIR timestamp
            # comparisons as reported in bug #332217. Also, fix
            # ownership since tar can change that too.
            os.utime(settings["WORKDIR"], None)
            _prepare_workdir(settings)
        elif self.phase == "install":
            out = io.StringIO()
            _post_src_install_write_metadata(settings)
            _post_src_install_uid_fix(settings, out)
            msg = out.getvalue()
            if msg:
                self.scheduler.output(msg, log_path=logfile)
        elif self.phase == "preinst":
            _preinst_bsdflags(settings)
        elif self.phase == "postinst":
            _postinst_bsdflags(settings)

        post_phase_cmds = _post_phase_cmds.get(self.phase)
        if post_phase_cmds is not None:
            if logfile is not None and self.phase in ("install",):
                # Log to a temporary file, since the code we are running
                # reads PORTAGE_LOG_FILE for QA checks, and we want to
                # avoid annoying "gzip: unexpected end of file" messages
                # when FEATURES=compress-build-logs is enabled.
                fd, logfile = tempfile.mkstemp()
                os.close(fd)
            post_phase = _PostPhaseCommands(
                background=self.background,
                commands=post_phase_cmds,
                elog=self._elog,
                fd_pipes=self.fd_pipes,
                logfile=logfile,
                phase=self.phase,
                scheduler=self.scheduler,
                settings=settings,
            )
            self._start_task(post_phase, self._post_phase_exit)
            return

        # this point is not reachable if there was a failure and
        # we returned for die_hooks above, so returncode must
        # indicate success (especially if ebuild_process.returncode
        # is unsuccessful and test-fail-continue came into play)
        self.returncode = os.EX_OK
        self._current_task = None
        self.wait()

    def _post_phase_exit(self, post_phase):

        self._assert_current(post_phase)

        log_path = None
        if self.settings.get("PORTAGE_BACKGROUND") != "subprocess":
            log_path = self.settings.get("PORTAGE_LOG_FILE")

        if post_phase.logfile is not None and post_phase.logfile != log_path:
            # We were logging to a temp file (see above), so append
            # temp file to main log and remove temp file.
            self._append_temp_log(post_phase.logfile, log_path)

        if self._final_exit(post_phase) != os.EX_OK:
            writemsg("!!! post %s failed; exiting.\n" % self.phase, noiselevel=-1)
            self._die_hooks()
            return

        self._current_task = None
        self.wait()
        return

    def _append_temp_log(self, temp_log, log_path):

        temp_file = open(
            _unicode_encode(temp_log, encoding=_encodings["fs"], errors="strict"), "rb"
        )

        log_file, log_file_real = self._open_log(log_path)

        for line in temp_file:
            log_file.write(line)

        temp_file.close()
        log_file.close()
        if log_file_real is not log_file:
            log_file_real.close()
        os.unlink(temp_log)

    def _open_log(self, log_path):

        f = open(
            _unicode_encode(log_path, encoding=_encodings["fs"], errors="strict"),
            mode="ab",
        )
        f_real = f

        if log_path.endswith(".gz"):
            f = gzip.GzipFile(filename="", mode="ab", fileobj=f)

        return (f, f_real)

    def _die_hooks(self):
        self.returncode = None
        phase = "die_hooks"
        die_hooks = MiscFunctionsProcess(
            background=self.background,
            commands=[phase],
            phase=phase,
            logfile=self._get_log_path(),
            fd_pipes=self.fd_pipes,
            scheduler=self.scheduler,
            settings=self.settings,
        )
        self._start_task(die_hooks, self._die_hooks_exit)

    def _die_hooks_exit(self, die_hooks):
        if (
            self.phase != "clean"
            and "noclean" not in self.settings.features
            and "fail-clean" in self.settings.features
        ):
            self._default_exit(die_hooks)
            self._fail_clean()
            return
        self._final_exit(die_hooks)
        self.returncode = 1
        self.wait()

    def _fail_clean(self):
        self.returncode = None
        portage.elog.elog_process(self.settings.mycpv, self.settings)
        phase = "clean"
        clean_phase = EbuildPhase(
            background=self.background,
            fd_pipes=self.fd_pipes,
            phase=phase,
            scheduler=self.scheduler,
            settings=self.settings,
        )
        self._start_task(clean_phase, self._fail_clean_exit)

    def _fail_clean_exit(self, clean_phase):
        self._final_exit(clean_phase)
        self.returncode = 1
        self.wait()

    async def _elog(self, elog_funcname, lines, background=None):
        if background is None:
            background = self.background
        out = io.StringIO()
        phase = self.phase
        elog_func = getattr(elog_messages, elog_funcname)
        global_havecolor = portage.output.havecolor
        try:
            portage.output.havecolor = self.settings.get(
                "NOCOLOR", "false"
            ).lower() in ("no", "false")
            for line in lines:
                elog_func(line, phase=phase, key=self.settings.mycpv, out=out)
        finally:
            portage.output.havecolor = global_havecolor
        msg = out.getvalue()
        if msg:
            build_logger = None
            try:
                log_file = None
                log_path = None
                if self.settings.get("PORTAGE_BACKGROUND") != "subprocess":
                    log_path = self.settings.get("PORTAGE_LOG_FILE")
                if log_path:
                    build_logger = BuildLogger(
                        env=self.settings.environ(),
                        log_path=log_path,
                        log_filter_file=self.settings.get(
                            "PORTAGE_LOG_FILTER_FILE_CMD"
                        ),
                        scheduler=self.scheduler,
                    )
                    build_logger.start()
                    _set_nonblocking(build_logger.stdin.fileno())
                    log_file = build_logger.stdin

                await self.scheduler.async_output(
                    msg, log_file=log_file, background=background
                )

                if build_logger is not None:
                    build_logger.stdin.close()
                    await build_logger.async_wait()
            except asyncio.CancelledError:
                if build_logger is not None:
                    build_logger.cancel()
                raise


class _PostPhaseCommands(CompositeTask):

    __slots__ = ("commands", "elog", "fd_pipes", "logfile", "phase", "settings")

    def _start(self):
        if isinstance(self.commands, list):
            cmds = [({}, self.commands)]
        else:
            cmds = list(self.commands)

        if "selinux" not in self.settings.features:
            cmds = [
                (kwargs, commands)
                for kwargs, commands in cmds
                if not kwargs.get("selinux_only")
            ]

        tasks = TaskSequence()
        for kwargs, commands in cmds:
            # Select args intended for MiscFunctionsProcess.
            kwargs = dict(
                (k, v) for k, v in kwargs.items() if k in ("ld_preload_sandbox",)
            )
            tasks.add(
                MiscFunctionsProcess(
                    background=self.background,
                    commands=commands,
                    fd_pipes=self.fd_pipes,
                    logfile=self.logfile,
                    phase=self.phase,
                    scheduler=self.scheduler,
                    settings=self.settings,
                    **kwargs
                )
            )

        self._start_task(tasks, self._commands_exit)

    def _commands_exit(self, task):

        if self._default_exit(task) != os.EX_OK:
            self._async_wait()
            return

        if self.phase == "install":
            out = io.StringIO()
            _post_src_install_soname_symlinks(self.settings, out)
            msg = out.getvalue()
            if msg:
                self.scheduler.output(
                    msg, log_path=self.settings.get("PORTAGE_LOG_FILE")
                )

            if "qa-unresolved-soname-deps" in self.settings.features:
                # This operates on REQUIRES metadata generated by the above function call.
                future = asyncio.ensure_future(
                    self._soname_deps_qa(), loop=self.scheduler
                )
                # If an unexpected exception occurs, then this will raise it.
                future.add_done_callback(
                    lambda future: future.cancelled() or future.result()
                )
                self._start_task(
                    AsyncTaskFuture(future=future), self._default_final_exit
                )
            else:
                self._default_final_exit(task)
        else:
            self._default_final_exit(task)

    async def _soname_deps_qa(self):

        vardb = QueryCommand.get_db()[self.settings["EROOT"]]["vartree"].dbapi

        all_provides = await self.scheduler.run_in_executor(
            ForkExecutor(loop=self.scheduler), _get_all_provides, vardb
        )

        unresolved = _get_unresolved_soname_deps(
            os.path.join(self.settings["PORTAGE_BUILDDIR"], "build-info"), all_provides
        )

        if unresolved:
            unresolved.sort()
            qa_msg = ["QA Notice: Unresolved soname dependencies:"]
            qa_msg.append("")
            qa_msg.extend(
                "\t%s: %s" % (filename, " ".join(sorted(soname_deps)))
                for filename, soname_deps in unresolved
            )
            qa_msg.append("")
            await self.elog("eqawarn", qa_msg)



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class RebuiltBinariesCase(TestCase):
    def testRebuiltBinaries(self):

        user_config = {
            "make.conf": ('FEATURES="binpkg-multi-instance"',),
        }

        binpkgs = (
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "1",
                    "BUILD_TIME": "1",
                },
            ),
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "2",
                    "BUILD_TIME": "2",
                },
            ),
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "3",
                    "BUILD_TIME": "3",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "1",
                    "BUILD_TIME": "1",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "2",
                    "BUILD_TIME": "2",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "3",
                    "BUILD_TIME": "3",
                },
            ),
        )

        installed = {
            "app-misc/A-1": {
                "EAPI": "5",
                "BUILD_ID": "1",
                "BUILD_TIME": "1",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "BUILD_ID": "2",
                "BUILD_TIME": "2",
            },
        }

        world = (
            "app-misc/A",
            "dev-libs/B",
        )

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--rebuilt-binaries": True,
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                ignore_mergelist_order=True,
                mergelist=["[binary]dev-libs/B-1-3", "[binary]app-misc/A-1-3"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                _user_config = user_config.copy()
                _user_config["make.conf"] += ('BINPKG_FORMAT="%s"' % binpkg_format,)
                playground = ResolverPlayground(
                    debug=False,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=_user_config,
                    world=world,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    # Disable debug so that cleanup works.
                    # playground.debug = False
                    playground.cleanup()



# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import re
from itertools import chain

from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.util import grabfile, write_atomic, ensure_dirs, normalize_path
from portage.const import USER_CONFIG_PATH, VCS_DIRS, WORLD_FILE, WORLD_SETS_FILE
from portage.localization import _
from portage.locks import lockfile, unlockfile
from portage import portage_gid
from portage._sets.base import PackageSet, EditablePackageSet
from portage._sets import SetConfigError, SETPREFIX, get_boolean
from portage.env.loaders import ItemFileLoader, KeyListFileLoader
from portage.env.validators import ValidAtomValidator
from portage import cpv_getkey

__all__ = [
    "StaticFileSet",
    "ConfigFileSet",
    "WorldSelectedSet",
    "WorldSelectedPackagesSet",
    "WorldSelectedSetsSet",
]


class StaticFileSet(EditablePackageSet):
    _operations = ["merge", "unmerge"]
    _repopath_match = re.compile(r".*\$\{repository:(?P<reponame>.+)\}.*")
    _repopath_sub = re.compile(r"\$\{repository:(?P<reponame>.+)\}")

    def __init__(self, filename, greedy=False, dbapi=None):
        super(StaticFileSet, self).__init__(allow_repo=True)
        self._filename = filename
        self._mtime = None
        self.description = "Package set loaded from file %s" % self._filename
        self.loader = ItemFileLoader(self._filename, self._validate)
        if greedy and not dbapi:
            self.errors.append(
                _(
                    "%s configured as greedy set, but no dbapi instance passed in constructor"
                )
                % self._filename
            )
            greedy = False
        self.greedy = greedy
        self.dbapi = dbapi

        metadata = grabfile(self._filename + ".metadata")
        key = None
        value = []
        for line in metadata:
            line = line.strip()
            if len(line) == 0 and key is not None:
                setattr(self, key, " ".join(value))
                key = None
            elif line[-1] == ":" and key is None:
                key = line[:-1].lower()
                value = []
            elif key is not None:
                value.append(line)
            else:
                pass
        else:
            if key is not None:
                setattr(self, key, " ".join(value))

    def _validate(self, atom):
        return bool(atom[:1] == SETPREFIX or ValidAtomValidator(atom, allow_repo=True))

    def write(self):
        write_atomic(
            self._filename,
            "".join(
                "%s\n" % (atom,) for atom in sorted(chain(self._atoms, self._nonatoms))
            ),
        )

    def load(self):
        try:
            mtime = os.stat(self._filename).st_mtime
        except (OSError, IOError):
            mtime = None
        if not self._loaded or self._mtime != mtime:
            try:
                data, errors = self.loader.load()
                for fname in errors:
                    for e in errors[fname]:
                        self.errors.append(fname + ": " + e)
            except EnvironmentError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e
                data = {}
            if self.greedy:
                atoms = []
                for a in data:
                    matches = self.dbapi.match(a)
                    for cpv in matches:
                        pkg = self.dbapi._pkg_str(cpv, None)
                        atoms.append("%s:%s" % (pkg.cp, pkg.slot))
                    # In addition to any installed slots, also try to pull
                    # in the latest new slot that may be available.
                    atoms.append(a)
            else:
                atoms = iter(data)
            self._setAtoms(atoms)
            self._mtime = mtime

    def singleBuilder(self, options, settings, trees):
        if not "filename" in options:
            raise SetConfigError(_("no filename specified"))
        greedy = get_boolean(options, "greedy", False)
        filename = options["filename"]
        # look for repository path variables
        match = self._repopath_match.match(filename)
        if match:
            try:
                filename = self._repopath_sub.sub(
                    trees["porttree"].dbapi.treemap[match.groupdict()["reponame"]],
                    filename,
                )
            except KeyError:
                raise SetConfigError(
                    _("Could not find repository '%s'") % match.groupdict()["reponame"]
                )
        return StaticFileSet(filename, greedy=greedy, dbapi=trees["vartree"].dbapi)

    singleBuilder = classmethod(singleBuilder)

    def multiBuilder(self, options, settings, trees):
        rValue = {}
        directory = options.get(
            "directory",
            os.path.join(settings["PORTAGE_CONFIGROOT"], USER_CONFIG_PATH, "sets"),
        )
        name_pattern = options.get("name_pattern", "${name}")
        if not "$name" in name_pattern and not "${name}" in name_pattern:
            raise SetConfigError(_("name_pattern doesn't include ${name} placeholder"))
        greedy = get_boolean(options, "greedy", False)
        # look for repository path variables
        match = self._repopath_match.match(directory)
        if match:
            try:
                directory = self._repopath_sub.sub(
                    trees["porttree"].dbapi.treemap[match.groupdict()["reponame"]],
                    directory,
                )
            except KeyError:
                raise SetConfigError(
                    _("Could not find repository '%s'") % match.groupdict()["reponame"]
                )

        try:
            directory = _unicode_decode(
                directory, encoding=_encodings["fs"], errors="strict"
            )
            # Now verify that we can also encode it.
            _unicode_encode(directory, encoding=_encodings["fs"], errors="strict")
        except UnicodeError:
            directory = _unicode_decode(
                directory, encoding=_encodings["fs"], errors="replace"
            )
            raise SetConfigError(
                _(
                    "Directory path contains invalid character(s) for encoding '%s': '%s'"
                )
                % (_encodings["fs"], directory)
            )

        vcs_dirs = [_unicode_encode(x, encoding=_encodings["fs"]) for x in VCS_DIRS]
        if os.path.isdir(directory):
            directory = normalize_path(directory)

            for parent, dirs, files in os.walk(directory):
                try:
                    parent = _unicode_decode(
                        parent, encoding=_encodings["fs"], errors="strict"
                    )
                except UnicodeDecodeError:
                    continue
                for d in dirs[:]:
                    if d in vcs_dirs or d.startswith(b".") or d.endswith(b"~"):
                        dirs.remove(d)
                for filename in files:
                    try:
                        filename = _unicode_decode(
                            filename, encoding=_encodings["fs"], errors="strict"
                        )
                    except UnicodeDecodeError:
                        continue
                    if filename.startswith(".") or filename.endswith("~"):
                        continue
                    if filename.endswith(".metadata"):
                        continue
                    filename = os.path.join(parent, filename)[1 + len(directory) :]
                    myname = name_pattern.replace("$name", filename)
                    myname = myname.replace("${name}", filename)
                    rValue[myname] = StaticFileSet(
                        os.path.join(directory, filename),
                        greedy=greedy,
                        dbapi=trees["vartree"].dbapi,
                    )
        return rValue

    multiBuilder = classmethod(multiBuilder)


class ConfigFileSet(PackageSet):
    def __init__(self, filename):
        super(ConfigFileSet, self).__init__()
        self._filename = filename
        self.description = "Package set generated from %s" % self._filename
        self.loader = KeyListFileLoader(self._filename, ValidAtomValidator)

    def load(self):
        data, errors = self.loader.load()
        self._setAtoms(iter(data))

    def singleBuilder(self, options, settings, trees):
        if not "filename" in options:
            raise SetConfigError(_("no filename specified"))
        return ConfigFileSet(options["filename"])

    singleBuilder = classmethod(singleBuilder)

    def multiBuilder(self, options, settings, trees):
        rValue = {}
        directory = options.get(
            "directory", os.path.join(settings["PORTAGE_CONFIGROOT"], USER_CONFIG_PATH)
        )
        name_pattern = options.get("name_pattern", "sets/package_$suffix")
        if not "$suffix" in name_pattern and not "${suffix}" in name_pattern:
            raise SetConfigError(_("name_pattern doesn't include $suffix placeholder"))
        for suffix in ["keywords", "use", "mask", "unmask"]:
            myname = name_pattern.replace("$suffix", suffix)
            myname = myname.replace("${suffix}", suffix)
            rValue[myname] = ConfigFileSet(os.path.join(directory, "package." + suffix))
        return rValue

    multiBuilder = classmethod(multiBuilder)


class WorldSelectedSet(EditablePackageSet):
    description = "Set of packages and subsets that were directly installed by the user"

    def __init__(self, eroot):
        super(WorldSelectedSet, self).__init__(allow_repo=True)
        self._pkgset = WorldSelectedPackagesSet(eroot)
        self._setset = WorldSelectedSetsSet(eroot)

    def write(self):
        self._pkgset._atoms = self._atoms.copy()
        self._pkgset.write()
        self._setset._nonatoms = self._nonatoms.copy()
        self._setset.write()

    def load(self):
        # Iterating over these sets does not force them to load if they
        # have been loaded previously.
        self._pkgset.load()
        self._setset.load()
        self._setAtoms(chain(self._pkgset, self._setset))

    def lock(self):
        self._pkgset.lock()
        self._setset.lock()

    def unlock(self):
        self._pkgset.unlock()
        self._setset.unlock()

    def cleanPackage(self, vardb, cpv):
        self._pkgset.cleanPackage(vardb, cpv)

    def singleBuilder(self, options, settings, trees):
        return WorldSelectedSet(settings["EROOT"])

    singleBuilder = classmethod(singleBuilder)


class WorldSelectedPackagesSet(EditablePackageSet):
    description = "Set of packages that were directly installed by the user"

    def __init__(self, eroot):
        super(WorldSelectedPackagesSet, self).__init__(allow_repo=True)
        self._lock = None
        self._filename = os.path.join(eroot, WORLD_FILE)
        self.loader = ItemFileLoader(self._filename, self._validate)
        self._mtime = None

    def _validate(self, atom):
        return ValidAtomValidator(atom, allow_repo=True)

    def write(self):
        write_atomic(self._filename, "".join(sorted("%s\n" % x for x in self._atoms)))

    def load(self):
        atoms = []
        atoms_changed = False
        try:
            mtime = os.stat(self._filename).st_mtime
        except (OSError, IOError):
            mtime = None
        if not self._loaded or self._mtime != mtime:
            try:
                data, errors = self.loader.load()
                for fname in errors:
                    for e in errors[fname]:
                        self.errors.append(fname + ": " + e)
            except EnvironmentError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e
                data = {}
            atoms = list(data)
            self._mtime = mtime
            atoms_changed = True
        else:
            atoms.extend(self._atoms)

        if atoms_changed:
            self._setAtoms(atoms)

    def _ensure_dirs(self):
        ensure_dirs(
            os.path.dirname(self._filename), gid=portage_gid, mode=0o2750, mask=0o2
        )

    def lock(self):
        if self._lock is not None:
            raise AssertionError("already locked")
        self._ensure_dirs()
        self._lock = lockfile(self._filename, wantnewlockfile=1)

    def unlock(self):
        if self._lock is None:
            raise AssertionError("not locked")
        unlockfile(self._lock)
        self._lock = None

    def cleanPackage(self, vardb, cpv):
        """
        Before calling this function you should call lock and load.
        After calling this function you should call unlock.
        """
        if not self._lock:
            raise AssertionError("cleanPackage needs the set to be locked")

        worldlist = list(self._atoms)
        mykey = cpv_getkey(cpv)
        newworldlist = []
        for x in worldlist:
            if x.cp == mykey:
                matches = vardb.match(x, use_cache=0)
                if not matches:
                    # zap our world entry
                    pass
                elif len(matches) == 1 and matches[0] == cpv:
                    # zap our world entry
                    pass
                else:
                    # others are around; keep it.
                    newworldlist.append(x)
            else:
                # this doesn't match the package we're unmerging; keep it.
                newworldlist.append(x)

        newworldlist.extend(self._nonatoms)
        self.replace(newworldlist)

    def singleBuilder(self, options, settings, trees):
        return WorldSelectedPackagesSet(settings["EROOT"])

    singleBuilder = classmethod(singleBuilder)


class WorldSelectedSetsSet(EditablePackageSet):
    description = "Set of sets that were directly installed by the user"

    def __init__(self, eroot):
        super(WorldSelectedSetsSet, self).__init__(allow_repo=True)
        self._lock = None
        self._filename = os.path.join(eroot, WORLD_SETS_FILE)
        self.loader = ItemFileLoader(self._filename, self._validate)
        self._mtime = None

    def _validate(self, setname):
        return setname.startswith(SETPREFIX)

    def write(self):
        write_atomic(
            self._filename, "".join(sorted("%s\n" % x for x in self._nonatoms))
        )

    def load(self):
        atoms_changed = False
        try:
            mtime = os.stat(self._filename).st_mtime
        except (OSError, IOError):
            mtime = None
        if not self._loaded or self._mtime != mtime:
            try:
                data, errors = self.loader.load()
                for fname in errors:
                    for e in errors[fname]:
                        self.errors.append(fname + ": " + e)
            except EnvironmentError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e
                data = {}
            nonatoms = list(data)
            self._mtime = mtime
            atoms_changed = True
        else:
            nonatoms = list(self._nonatoms)

        if atoms_changed:
            self._setAtoms(nonatoms)

    def lock(self):
        if self._lock is not None:
            raise AssertionError("already locked")
        self._lock = lockfile(self._filename, wantnewlockfile=1)

    def unlock(self):
        if self._lock is None:
            raise AssertionError("not locked")
        unlockfile(self._lock)
        self._lock = None

    def singleBuilder(self, options, settings, trees):
        return WorldSelectedSetsSet(settings["EROOT"])

    singleBuilder = classmethod(singleBuilder)
